#ifndef __DEF_MATRIXOPS_H
#define __DEF_MATRIXOPS_H

#include "helper.h"

// helper functions

template <class T>
size_t size( const vector< T >& v )
{
	return v.size();
}


template <class T>
vector< vector<T> > invert( const vector< vector<T> >& cov, bool &problem )
{
	vector< vector<T> > invcov( cov.size() );
	for (size_t i = 0; i < cov.size(); i++)
		invcov[i].resize( cov.size(), 0.0 );

	char UPLO = 'L';
	integer N = (integer)cov.size(); 
	double *COV = new double[ cov.size()*(cov.size()+1) / 2 ];
	double *WORK = new double[ cov.size() ];
	integer *IPIV = new integer[ cov.size() ];
	integer INFO;

	for (size_t i = 0; i < cov.size(); i++)
		for (size_t j = 0; j <= i; j++)
			COV[i + j * (2 * cov.size() - j - 1) / 2] = double( cov[i][j] );
	dsptrf_(&UPLO, &N, COV, IPIV, &INFO );
	if (INFO != 0)
	{
		std::cout << "invert failed in dsptrf_ with exit code" << INFO << std::endl;
		for (size_t i = 0; i < cov.size(); i++)
		{
			invcov[i][i] = 1.0 / cov[i][i];
		}		
		problem = true;
		return invcov;
	}
	dsptri_(&UPLO, &N, COV, IPIV, WORK, &INFO );
	if (INFO != 0)
	{
		std::cout << "invert failed in dsptri_ with exit code" << INFO << std::endl;
		for (size_t i = 0; i < cov.size(); i++)
		{
			invcov[i][i] = 1.0 / cov[i][i];
		}		
		return invcov;
	}

	for (size_t i = 0; i < cov.size(); i++)
	{
		for (size_t j = 0; j <= i; j++)
		{
			invcov[i][j] = T( COV[i + j * (2 * cov.size() - j - 1) / 2] );
			invcov[j][i] = invcov[i][j];
		}
	}

	delete[] COV;
	delete[] WORK;
	delete[] IPIV;
	return invcov;
}

template <class T>
double determinant( const vector< vector<T> >& cov, size_t& rank )
{
	double determinant = 1.0;
	char JOBZ = 'N';
	char UPLO = 'L';

	integer N = cov.size(); 
	double *EIG = new double[ cov.size() ];
	double *WORK = new double[ (cov.size()+2) * cov.size() ];
	integer LWORK = (cov.size()+2) * cov.size();
	double *COV = new double[ cov.size()*cov.size() ];
	integer INFO;
	for (size_t i = 0; i < cov.size(); i++)
		for (size_t j = 0; j < cov.size(); j++)
			COV[i + cov.size()*j] = double( cov[i][j] );
	dsyev_( &JOBZ, &UPLO, &N, COV, &N, EIG, WORK, &LWORK, &INFO );

	rank = 0;
	for (size_t i = 0; i < cov.size(); i++)
	{
		if ( EIG[i] > EPSILOND )
			rank++;
		determinant *= T( max(EIG[i], T(0.0)) );
	}

	delete[] EIG;
	delete[] WORK;
	delete[] COV;
	return determinant;
}

template <class T>
double determinant( const vector< vector<T> >& cov )
{
	size_t rank;
	return determinant<T>(cov, rank);
}


template <class T>
vector<T> replaceElementsOfYAccordingToX( const vector< T >& x, const vector< vector< T > >& y )
{
	vector<T> yx( x.size() );
	for (size_t d = 0; d < x.size(); d++)
	{
		yx[d] = y[d][ x[d] ];
	}
	return yx;
}

template <class T>
vector<T> subvector( const vector< T >& x, size_t d1 )
{
	vector<T> xx( x.size() - d1 );
	size_t s = 0;
	for (size_t d = d1; d < x.size(); d++, s++)
	{
		xx[s] = x[d];
	}
	return xx;
}

template <class T>
vector<T> subvector( const vector< T >& x, size_t d1, size_t d2 )
{
	vector<T> xx( d2 - d1 + 1 );
	size_t s = 0;
	for (size_t d = d1; d <= d2; d++, s++)
	{
		xx[s] = x[d];
	}
	return xx;
}


template <class T>
vector< vector<T> > subsqmatrix( const vector< vector< T > >& x, size_t d1, size_t d2 )
{
	vector< vector<T> > xx( d2 - d1 + 1 );
	size_t s1 = 0;
	for (size_t da = d1; da <= d2; da++, s1++)
	{
		size_t s2 = 0;
		xx[s1].resize( d2 - d1 + 1 );
		for (size_t db = d1; db <= d2; db++, s2++)
		{
			xx[s1][s2] = x[da][db];
		}
	}
	return xx;
}



template <class T>
inline vector< vector<T> > mult( vector< vector<T> >& x, vector< vector<T> >& y )
{
	size_t M = x.size();
	size_t N = x[0].size();
	size_t L = y.size();
	if (N != L )
	{
		std::cout << "mult: number of columns in x is not equal to number of rows in y" << std::endl;
		std::cout << "x is " << M << "x" <<  N << std::endl;
		std::cout << "y is " << L << "x" <<  y[0].size() << std::endl;
		exit(-1);
	}
	size_t K = y[0].size();
	vector< vector<T> > z( M );
	for (size_t i = 0; i < M; i++) 
	{
		z[i].resize( K, T(0) );
		for (size_t j = 0; j < K; j++) 
		{
			for (size_t r = 0; r < N; r++) 
			{
				z[i][j] += x[i][r] * y[r][j];
			}
		}
	}
	return z;
}



template <class T1, class T2>
inline vector< vector<T1> > mult1( vector< vector<T1> >& x, vector< vector<T2> >& y )
{
	size_t M = x.size();
	size_t N = x[0].size();
	size_t L = y.size();
	if (N != L )
	{
		std::cout << "mult1: number of columns in x is not equal to number of rows in y" << std::endl;
		std::cout << "x is " << M << "x" <<  N << std::endl;
		std::cout << "y is " << L << "x" <<  y[0].size() << std::endl;
		exit(-1);
	}
	size_t K = y[0].size();
	vector< vector<T1> > z( M );
	for (size_t i = 0; i < M; i++) 
	{
		z[i].resize( K, T1(0) );
		for (size_t j = 0; j < K; j++) 
		{
			for (size_t r = 0; r < N; r++) 
			{
				z[i][j] += x[i][r] * (T1)y[r][j];
			}
		}
	}
	return z;
}

template <class T1, class T2>
inline vector< vector<T2> > mult2( vector< vector<T1> >& x, vector< vector<T2> >& y )
{
	size_t M = x.size();
	size_t N = x[0].size();
	size_t L = y.size();
	if (N != L )
	{
		std::cout << "mult2: number of columns in x is not equal to number of rows in y" << std::endl;
		std::cout << "x is " << M << "x" <<  N << std::endl;
		std::cout << "y is " << L << "x" <<  y[0].size() << std::endl;
		exit(-1);
	}
	size_t K = y[0].size();
	vector< vector<T2> > z( M );
	for (size_t i = 0; i < M; i++) 
	{
		z[i].resize( K, T2(0) );
		for (size_t j = 0; j < K; j++) 
		{
			for (size_t r = 0; r < N; r++) 
			{
				z[i][j] += (T2)x[i][r] * y[r][j];
			}
		}
	}
	return z;
}


template <class T>
inline vector<T> mult( vector< vector<T> >& x, vector< T >& y )
{
	size_t M = x.size();
	size_t N = x[0].size();
	size_t L = y.size();
	if (N != L )
	{
		std::cout << "mult matrix-vector: number of columns in x is not equal to number of rows in y" << std::endl;
		std::cout << "x is " << M << "x" <<  N << std::endl;
		std::cout << "y is " << L << "x1" << std::endl;
		assert(false);
		exit(-1);
	}
	vector< T > z( M, 0 );
	for (size_t i = 0; i < M; i++) 
	{
		for (size_t r = 0; r < N; r++) 
		{
			z[i] += x[i][r] * y[r];
		}
	}
	return z;
}

template <class T>
inline vector<T> mult( vector<T>& x, T y )
{
	vector<T> z( x.size(), T(0) );
	for (size_t d = 0; d < x.size(); d++)
		z[d] = x[d] * y;
	return z;
}

template <class T>
inline vector< vector<T> > mult( vector< vector<T> >& x, T y )
{
	vector< vector<T> > z( x.size() );
	for (size_t d = 0; d < x.size(); d++)
		z[d].resize( x[d].size(), 0.0 );
	for (size_t d = 0; d < x.size(); d++)
		for (size_t d2 = 0; d2 < x[d].size(); d2++)
			z[d][d2] = x[d][d2] * y;
	return z;
}


template <class T>
inline vector<T> mult( vector<T>& x, vector<T>& y )
{
	vector<T> z( x.size(), T(0) );
	for (size_t d = 0; d < x.size(); d++)
		z[d] = x[d] * y[d];
	return z;
}

template <class T1, class T2>
inline vector<T1> mult1( vector<T1>& x, vector<T2>& y )
{
	vector<T1> z( x.size(), T1(0) );
	for (size_t d = 0; d < x.size(); d++)
		z[d] = x[d] * (T1)y[d];
	return z;
}

template <class T1, class T2>
inline vector<T2> mult2( vector<T1>& x, vector<T2>& y )
{
	vector<T2> z( x.size(), T2(0) );
	for (size_t d = 0; d < x.size(); d++)
		z[d] = (T2)x[d] * y[d];
	return z;
}


template <class T>
inline vector<T> add( vector<T>& x, vector<T>& y )
{
	vector<T> z( x.size(), T(0) );
	for (size_t d = 0; d < x.size(); d++)
		z[d] = x[d] + y[d];
	return z;
}

template <class T>
inline vector<T> add( vector<T>& x, T y )
{
	vector<T> z( x.size(), T(0) );
	for (size_t d = 0; d < x.size(); d++)
		z[d] = x[d] + y;
	return z;
}


template <class T>
inline vector< vector<T> > add( vector< vector<T> >& x, vector< vector<T> >& y )
{
	vector< vector<T> > z( x.size() );
	for (size_t d = 0; d < x.size(); d++)
		z[d].resize( x[d].size(), T(0)  );
	for (size_t d = 0; d < x.size(); d++)
		for (size_t d2 = 0; d2 < x[d].size(); d2++)
			z[d][d2] = x[d][d2] + y[d][d2];
	return z;
}

template <class T>
inline vector< vector<T> > regularize( vector< vector<T> >& x, T reg )
{
	vector< vector<T> > z( x.size() );
	for (size_t d = 0; d < x.size(); d++)
		z[d].resize( x[d].size(), T(0)  );

	for (size_t d = 0; d < x.size(); d++)
		for (size_t d2 = 0; d2 < x[d].size(); d2++)
			z[d][d2] = x[d][d2] * (1 - reg);

	for (size_t d = 0; d < x.size(); d++)
		z[d][d] = x[d][d] + reg;
	return z;
}


template <class T>
inline bool equals( vector< T >& x, vector< T >& y )
{
	for (size_t d = 0; d < x.size(); d++)
		if ( x[d] != y[d] )
			return false;
	return true;
}


template <class T>
inline bool equals( vector< vector<T> >& x, vector< vector<T> >& y )
{
	for (size_t d = 0; d < x.size(); d++)
		for (size_t d2 = 0; d2 < x[d].size(); d2++)
			if ( x[d][d2] != y[d][d2] )
				return false;
	return true;
}


template <class T>
inline vector<T> subtract( vector<T>& x, vector<T>& y )
{
	vector<T> z( x.size(), T(0) );
	for (size_t d = 0; d < x.size(); d++)
		z[d] = x[d] - y[d];
	return z;
}

template <class T1, class T2>
inline vector<T1> subtract1( vector<T1>& x, vector<T2>& y )
{
	vector<T1> z( x.size(), T1(0) );
	for (size_t d = 0; d < x.size(); d++)
		z[d] = x[d] - (T1)y[d];
	return z;
}

template <class T1, class T2>
inline vector<T2> subtract2( vector<T1>& x, vector<T2>& y )
{
	vector<T2> z( x.size(), T2(0) );
	for (size_t d = 0; d < x.size(); d++)
		z[d] = (T2)x[d] - y[d];
	return z;
}




template <class T>
inline vector< vector<T> > subtract( vector< vector<T> >& x, vector< vector<T> >& y )
{
	vector< vector<T> > z( x.size() );
	for (size_t d = 0; d < x.size(); d++)
		z[d].resize( x[d].size(), T(0)  );
	for (size_t d = 0; d < x.size(); d++)
		for (size_t d2 = 0; d2 < x[d].size(); d2++)
			z[d][d2] = x[d][d2] - y[d][d2];
	return z;
}

template <class T1, class T2>
inline vector< vector<T1> > subtract1( vector< vector<T1> >& x, vector< vector<T2> >& y )
{
	vector< vector<T1> > z( x.size() );
	for (size_t d = 0; d < x.size(); d++)
		z[d].resize( x[d].size(), T1(0) );
	for (size_t d = 0; d < x.size(); d++)
		for (size_t d2 = 0; d2 < x[d].size(); d2++)
			z[d][d2] = x[d][d2] - (T1)y[d][d2];
	return z;
}

template <class T1, class T2>
inline vector< vector<T2> > subtract2( vector< vector<T1> >& x, vector< vector<T2> >& y )
{
	vector< vector<T2> > z( x.size() );
	for (size_t d = 0; d < x.size(); d++)
		z[d].resize( x[d].size(), T2(0) );
	for (size_t d = 0; d < x.size(); d++)
		for (size_t d2 = 0; d2 < x[d].size(); d2++)
			z[d][d2] = (T2)x[d][d2] - y[d][d2];
	return z;
}

template <class T>
inline vector< vector<T> > transpose( vector< vector<T> >& x )
{
	vector< vector<T> > z( x[0].size() );
	for (size_t d = 0; d < x[0].size(); d++)
		z[d].resize( x.size(), T(0) );
	for (size_t d = 0; d < x[0].size(); d++)
		for (size_t d2 = 0; d2 < x.size(); d2++)
			z[d][d2] = x[d2][d];
	return z;
}

template <class T>
inline void selfadd( vector<T>& x, vector<T> y )
{
	for (size_t d = 0; d < x.size(); d++)
		x[d] += y[d];
}

template <class T>
inline T sum( vector<T>& x )
{
	T sum = 0.0;
	for (size_t d = 0; d < x.size(); d++)
		sum += x[d];
	return sum;
}


template <class T>
inline T sum( vector< vector<T> >& x )
{
	T sum = 0.0;
	for (size_t d = 0; d < x.size(); d++)
		for (size_t d2 = 0; d2 < x[d].size(); d2++)
			sum += x[d][d2];
	return sum;
}

template <class T>
inline T sumabs( vector< vector<T> >& x )
{
	if ( x.empty() )
		return std::numeric_limits<T>::max();
	T sum = 0.0;
	for (size_t d = 0; d < x.size(); d++)
		for (size_t d2 = 0; d2 < x[d].size(); d2++)
			sum += abs( x[d][d2] );
	return sum;
}



template <class T>
inline void selfsubtract( vector<T>& x, vector<T> y )
{
	for (size_t d = 0; d < x.size(); d++)
		x[d] -= y[d];
}

template <class T>
inline void selfsubtract( vector<T>& x, T y )
{
	for (size_t d = 0; d < x.size(); d++)
		x[d] -= y;
}

template <class T>
inline void selfsubtract_butkeepitmax( vector<T>& x, T y, T m )
{
	for (size_t d = 0; d < x.size(); d++)
	{
		x[d] -= y;
		x[d] = max( x[d], m );
	}
}

template <class T, class T2>
inline vector<T2> convert( vector<T>& x )
{
	vector<T2> xx( x.size() );
	for (size_t d = 0; d < x.size(); d++)
	{
		xx[d] = T2( x[d] );
	}
	return xx;
}

template <class T>
inline void selfdiv( vector<T>& x, T y )
{
	for (size_t d = 0; d < x.size(); d++)
		x[d] /= y;
}


template <class T>
inline void selfadd( vector< vector<T> >& x, vector< vector<T> > y )
{
	for (size_t d = 0; d < x.size(); d++)
		for (size_t d2 = 0; d2 < x[d].size(); d2++)
			x[d][d2] += y[d][d2];
}

template <class T>
inline void selfdiv( vector< vector<T> >& x, T y )
{
	for (size_t d = 0; d < x.size(); d++)
		for (size_t d2 = 0; d2 < x[d].size(); d2++)
			x[d][d2] /=  y;
}


template <class T>
inline T dist2( vector<T>& x, vector<T>& y )
{
	T dist2 = 0.0;
	for (size_t d = 0; d < x.size(); d++)
		dist2 += (y[d] - x[d])*(y[d] - x[d]);
	return dist2;
}


template <class T>
inline T dist( vector<T>& x, vector<T>& y )
{
	T dist2 = 0.0;
	for (size_t d = 0; d < x.size(); d++)
		dist2 += (y[d] - x[d])*(y[d] - x[d]);
	return sqrt(dist2);
}

template <class T1, class T2>
inline T1 dist( vector<T1>& x, vector<T2>& y )
{
	T1 dist2 = 0.0;
	for (size_t d = 0; d < x.size(); d++)
		dist2 += ((T1)y[d] - x[d])*((T1)y[d] - x[d]);
	return sqrt(dist2);
}


template <class T1, class T2>
inline T1 dist2( vector<T1>& x, vector<T2>& y )
{
	T1 dist2 = 0.0;
	for (size_t d = 0; d < x.size(); d++)
		dist2 += ((T1)y[d] - x[d])*((T1)y[d] - x[d]);
	return dist2;
}

template <class T>
inline vector< vector<double> > covfromvector( vector<T>& x, T regul )
{
	vector< vector<double> > c( x.size() );
	for (size_t d = 0; d < x.size(); d++)
		c[d].resize( x.size(), 0.0 );	

	for (size_t d = 0; d < x.size(); d++)
		for (size_t d2 = 0; d2 < x.size(); d2++)
			c[d][d2] = x[d] * x[d2] * (1 - regul);

	for (size_t d = 0; d < x.size(); d++)
		c[d][d] += regul;										// regularization

	return c;
}

template <class T>
inline vector< vector<double> > covfromvector( vector<T>& x, vector<T>& y )
{
	vector< vector<double> > c( x.size() );
	for (size_t d = 0; d < x.size(); d++)
		c[d].resize( y.size(), 0.0 );	

	for (size_t d = 0; d < x.size(); d++)
		for (size_t d2 = 0; d2 < y.size(); d2++)
			c[d][d2] = x[d] * y[d2];

	return c;
}

template <class T>
inline void concatenateUniqueVectors(vector<T>& v1, vector<T>& v2) 
{
	for (size_t j = 0; j < v2.size(); j++)
	{
		bool found = false;
		for (size_t i = 0; i < v1.size(); i++)
		{
			if ( v2[j] == v1[i] )
				found = true;
		}
		if (!found)
			v1.push_back( v2[j] );
	}
}

//
//template <class T>
//inline vector<T> concatenateVectors(vector<T>& v1, vector<T> v2) 
//{
//	v1.insert( v1.end(), v2.begin(), v2.end() );
//	return v1;
//}
//
//
//template <class T, class T2>
//inline vector<T> concatenateVectors(vector<T>& v1, vector<T2> v2) 
//{
//	for (size_t d = 0; d < v2.size(); d++)
//	{
//		v1.push_back( (T)v2[d] );
//	}
//	return v1;
//}


template <class T>
inline vector<T> concatenateVectors(vector<T> v1, vector<T> v2) 
{
	v1.insert( v1.end(), v2.begin(), v2.end() );
	return v1;
}


template <class T, class T2>
inline vector<T> concatenateVectors(vector<T> v1, vector<T2> v2) 
{
	for (size_t d = 0; d < v2.size(); d++)
	{
		v1.push_back( (T)v2[d] );
	}
	return v1;
}

template <class T, class T2>
inline vector<T> concatenateVectors(vector<T> v1, vector< vector<T2> > v2) 
{
	for (size_t d = 0; d < v2.size(); d++)
	{
		for (size_t d2 = 0; d2 < v2[d].size(); d2++)
		{
			v1.push_back( (T)v2[d][d2] );
		}
	}
	return v1;
}


template <class T>
inline std::vector<size_t> sortVectorIndices(vector<T>& V) {
	vector<size_t> VI(V.size());
	for (size_t i = 0; i < V.size(); i++) {
		VI[i] = i;
	}		
	std::sort( VI.begin(), VI.end(), IndexCmp<vector<T>&>(V) );
	return VI;
};




template <class T>
inline T maximum(T* v, size_t size) {
	T result = boost::numeric::bounds<T>::lowest();
	for (size_t i = 0; i < size; i++) {
		if (v[i] > result) {
			result = v[i];
		}
	}
	return result;
}

template <class T>
inline T maximum(vector<T>& v) {
	T result = boost::numeric::bounds<T>::lowest();
	for (int i = 0; i < v.size(); i++) {
		if (v[i] > result) {
			result = v[i];
		}
	}
	return result;
}


template <class T>
inline T maximumabs(vector<T>& v) 
{
	T result = boost::numeric::bounds<T>::lowest();
	for (int i = 0; i < v.size(); i++) 
	{
		if (abs<T>(v[i]) > result) {
			result = abs<T>(v[i]);
		}
	}
	return result;
}


template <class T>
inline size_t maximumId(vector<T>& v) {
  size_t id = 0;
	T result = boost::numeric::bounds<T>::lowest();
	for (size_t i = 0; i < v.size(); i++) {
		if (v[i] > result) {
			result = v[i];
      id = i;
		}
	}
	return id;
}

template <class T>
inline vector<size_t> maximumId_perrow(vector< vector<T> >& v) {
	vector<size_t> maxid( v.size() );
	for (size_t d = 0; d < v.size(); d++)
		maxid[d] = maximumId( v[d] );
	return maxid;
}

template <class T>
inline size_t maximumId_percolumn(vector< vector<T> >& v) {
	vector< vector<T> > vt = transpose(v);
	vector<size_t> maxid( vt.size() );
	for (size_t d = 0; d < vt.size(); d++)
		maxid[d] = maximumId( vt[d] );
	return maxid;
}

template <class T>
inline T minimum(vector<T>& v) {
	T result = std::numeric_limits<T>::max(); 
	for (int i = 0; i < v.size(); i++) {
		if (v[i] < result) {
			result = v[i];
		}
	}
	return result;
}

template <class T>
inline size_t minimumId(vector<T>& v) {
  size_t id = 0;
	T result = std::numeric_limits<T>::max(); 
	for (size_t i = 0; i < v.size(); i++) {
		if (v[i] < result) {
			result = v[i];
      id = i;
		}
	}
	return id;
}


template <class T>
inline T mean(vector<T>& v) {
	T result = 0;
	for (int i = 0; i < v.size(); i++) {
		result += v[i];
	}
	result /= (T)v.size();
	return result;
}


template <class T1, class T2>
inline void averageWithExistingVector(vector<T1>& v, vector<T2>& w) {
	if (v.empty())
	{
		v.resize( w.size() );
		for (size_t i = 0; i < v.size(); i++)
			v[i] = (T1)w[i];
	}
	else
	{
		for (size_t i = 0; i < v.size(); i++)
			v[i] = (v[i] + (T1)w[i]) / T1(2);
	}
}


template <class T>
inline T weightedmean(vector<T>& v, vector<T>& w) {
	T result = 0;
	T sumw = 0;
	for (int i = 0; i < v.size(); i++) {
		result += v[i] * w[i];
		sumw += w[i];
	}

	result /= sumw;
	return result;
}

template <class T>
inline T median(vector<T> v) {
	std::sort(v.begin(), v.end());
	if ( v.size() % 2 == 0 ) {
		return ( v[ v.size() / 2 ] +  v[ (v.size() / 2) - 1 ] ) / 2.0f;
	} else {
		return v[ v.size() / 2 ];
	}
}

template <class T>
inline T weightedmedian(vector<T>& v, vector<T>& w) {
	vector<size_t> VI = sortVectorIndices(v);
	double sumw = 0.0f, csumw = 0.0f;
	int i;
	for (i = 0; i < w.size(); i++) {
		sumw += (double)w[i];
	}
	for (i = 0; i < w.size(); i++) {
		csumw += (double)w[ VI[i] ] / (double)sumw;
		if (csumw >= .5f)
			break;
	}
	return v[ VI[i] ];
}


template <class T>
inline T mean2(vector<T>& v) {
	T result = 0;
	for (int i = 0; i < v.size(); i++) {
		result += v[i] * v[i];
	}
	result /= (T)v.size();
	return result;
}

template <class T>
inline T mean4(vector<T>& v) {
	T result = 0;
	for (int i = 0; i < v.size(); i++) {
		result += v[i] * v[i] * v[i] * v[i];
	}
	result /= (T)v.size();
	return result;
}

template <class T>
inline T meansqrt(vector<T>& v) {
	T result = 0;
	for (int i = 0; i < v.size(); i++) {
		result += sqrt(v[i]);
	}
	result /= (T)v.size();
	return result;
}

template <class T>
inline T meansqrtsqrt(vector<T>& v) {
	T result = 0;
	for (int i = 0; i < v.size(); i++) {
		result += sqrt(sqrt(v[i]));
	}
	result /= (T)v.size();
	return result;
}


template <class T>
inline T variance(vector<T>& v, T meanValue) {
	T result = 0;
	for (int i = 0; i < v.size(); i++) {
		result += (v[i] - meanValue) * (v[i] - meanValue);
	}
	result /= (T)v.size();
	return result;
}

template <class T>
inline T percentile(vector<T> v, float k) {
	std::sort(v.begin(), v.end());
	if ( k == 0.0f )
		return v[0];
	else if ( k == 1.0f ) 
		return v[ v.size()-1 ];
	else
		return v[ (int)floor( k * (float)v.size() ) ];
}


template <class T>
inline T robust_variance(vector<T>& v, T meanValue) {
	T result = 0;
  T v1 = percentile(v, .1f);
  T v2 = percentile(v, .9f);
	for (size_t i = 0; i < v.size(); i++) {
    if ( v[i] > v1 && v[i] < v2 && fabs(v[i]) < FLT_MAX && v[i] == v[i]  )
  		result += (v[i] - meanValue) * (v[i] - meanValue);
	}
	result /= (T)v.size();
	return result;
}


template <class T>
inline T skewness(vector<T>& v, T meanValue, T varianceValue) {
	T result = 0;
	for (int i = 0; i < v.size(); i++) {
		result += (v[i] - meanValue) * (v[i] - meanValue) * (v[i] - meanValue);
	}
	result /= (T)v.size();
	result /= sqrt( varianceValue*varianceValue*varianceValue + EPSILOND);
	return result;
}

template <class T>
inline T kurtosis(vector<T>& v, T meanValue, T varianceValue) {
	T result = 0;
	for (int i = 0; i < v.size(); i++) {
		result += (v[i] - meanValue) * (v[i] - meanValue) * (v[i] - meanValue) * (v[i] - meanValue);
	}
	result /= (T)v.size();
	result /= ( varianceValue*varianceValue  + EPSILOND);
	return result;
}

template <class T>
inline std::vector<T> unique(vector<T>& v) {
	std::vector<T> ret;
	ret.push_back( v[0] );
	for (int i = 1; i < v.size(); i++) {
		bool isunique = true;
		for (int j = 0; j < ret.size(); j++) {
			if (v[i] == ret[j]) {
				isunique = false;
				break;
			}
		}
		if (isunique == true) {
			ret.push_back( v[i] );
		}
	}

	return ret;
}



template <class T>
inline vector<T> removePositiveOutliers(vector<T>& v, float percentile) {
	vector<T> result( v.size() );
	for (int i = 0; i < v.size(); i++) {
		result[i] = fabs(v[i]);
	}
	std::sort( result.begin(), result.end() );
	float outlierValue = result[ (size_t)floor(percentile*(float)v.size()) ];
	for (int i = 0; i < v.size(); i++) {
		if (fabs(v[i]) < outlierValue) {
			result[i] = v[i];
		} else {
			result[i] = sgn(v[i]) * outlierValue;
		}
	}
	return result;
}

template <class T>
inline vector<T> removeOutliers(vector<T>& v, float percentile) {
	vector<T> result( v.size() );
	for (int i = 0; i < v.size(); i++) {
		result[i] = v[i];
	}
	std::sort( result.begin(), result.end() );
	float outlierValueMax = result[ (size_t)floor(percentile*(float)v.size()) ];
	float outlierValueMin = result[ (size_t)floor((1.0f-percentile)*(float)v.size()) ];
	for (int i = 0; i < v.size(); i++) {
		if (v[i] < outlierValueMin) {
			result[i] = outlierValueMin;
		} else if (v[i] > outlierValueMax) {
			result[i] = outlierValueMax;
		} else {
			result[i] = v[i];
		}
	}
	return result;
}

template <class T>
inline vector<T> removeOutliers(vector<T>& v, float percentilePositive, float percentileNegative) {
	vector<T> result( v.size() );
	for (int i = 0; i < v.size(); i++) {
		result[i] = v[i];
	}
	std::sort( result.begin(), result.end() );
	float outlierValueMax = result[ (size_t)floor(percentilePositive*(float)v.size()) ];
	float outlierValueMin = result[ (size_t)floor(percentileNegative*(float)v.size()) ];
	for (int i = 0; i < v.size(); i++) {
		if (v[i] < outlierValueMin) {
			result[i] = outlierValueMin;
		} else if (v[i] > outlierValueMax) {
			result[i] = outlierValueMax;
		} else {
			result[i] = v[i];
		}
	}
	return result;
}

template <class T>
inline vector<float> hist(vector<T>& v, int numbins, float minimum_v = T(0), float maximum_v = T(1) ) {
	vector<float> values( numbins, 0.0f );
	float lim = (float)minimum_v;
	float spacing = ((float)maximum_v - (float)minimum_v) / float(numbins);
	vector<float> limits;
	for (size_t b = 0; b < numbins; b++)
	{
		limits.push_back( lim );
		lim += spacing;
	}

	for (size_t i = 0; i < v.size(); i++)
	{
		size_t b = 0;
		for (b = 0; b < numbins-1; b++)
			if ((float)v[i] >= limits[b] && (float)v[i] < limits[b+1])
				break;
		values[b] += 1.0f;
	}

	T norm = sum( values );
	for (size_t i = 0; i < values.size(); i++)
		values[i] /= norm;
	return values;
}


// iostream operators
template <class T>
static inline std::ostream &operator << (std::ostream &os, const vector<T> &v)
{
	os << v.size() << " ";
	for (size_t i = 0; i < v.size(); i++)
		os << v[i] << " ";
	return os << std::endl;
}

template <class T>
static inline std::ostream &operator << (std::ostream &os, const vector< vector<T> > &v)
{
	os << v.size() << " ";
	for (size_t i = 0; i < v.size(); i++)
	{
		os << v[i].size() << " ";
		for (size_t j = 0; j < v[i].size(); j++)
			os << v[i][j] << " ";
		os << std::endl;
	}
	return os;
}



template <class T>
static inline std::istream &operator >> (std::istream &is, vector<T> &v)
{
	size_t s;
	is >> s;
	v.resize( s );
	for (size_t i = 0; i < s; i++)
		is >> v[i];
	return is;
}

template <class T>
static inline std::istream &operator >> (std::istream &is, vector< vector<T> > &v)
{
	size_t s;
	is >> s;
	v.resize( s );
	for (size_t i = 0; i < v.size(); i++)
	{
		is >> s;
		v[i].resize(s);
		for (size_t j = 0; j < v[i].size(); j++)
			is >> v[i][j];
	}
	return is;
}


inline double* covar(double* D, size_t N) {
	double *CD = new double[9];

	CD[0] = 0.0f;
	for (size_t r = 0; r < N; r++)
		CD[0]  += D[r*3] * D[r*3];
	for (size_t j = 0; j <= 1; j++) {
		CD[1 + j*3] = 0.0f;
		for (size_t r = 0; r < N; r++) {
			CD[1 + j*3] +=  D[1 + r*3] * D[j + r*3];
		}
	}
	for (size_t j = 0; j <= 2; j++) {
		CD[2 + j*3] = 0.0f;
		for (size_t r = 0; r < N; r++) {
			CD[2 + j*3] +=  D[2 + r*3] * D[j + r*3];
		}
	}
	CD[3] = CD[1];
	CD[6] = CD[2];
	CD[7] = CD[5];

	return CD;
}


inline vector<size_t> fromOffsetToVector(size_t offset, vector<size_t> parent_domain_sizes)
{
	vector<size_t> x( parent_domain_sizes.size() );
	for (int k = 0; k < x.size(); k++)
	{
		int prod = 1;
		for (int l = k+1; l < x.size(); l++)
			prod *= parent_domain_sizes[l];
		x[k] = offset / prod;
		offset = offset % prod;
	}
//	return offset;
	return x;
}


static bool histTest() {
	std::cout << "Testing hist functionality...";

	//vector< size_t > X;
	//X.push_back(3);
	//X.push_back(4);
	//X.push_back(2);
	//X.push_back(1);
	//for (size_t d = 0; d < 3*4*2*1; d++)
	//	std::cout << fromOffsetToVector(d, X ) << std::endl;
	//system("pause");

	vector< float > votes;
	votes.push_back( 0.4f );  // 3
	votes.push_back( 0.33f ); // 3
	votes.push_back( 0.2f );  // 2
	votes.push_back( 0.5f );  // 4
	votes.push_back( 0.7f );  // 5
	votes.push_back( 0.8f );  // 6
	votes.push_back( 1.0f );  // 7
	votes.push_back( 1.0f );  // 7
	votes.push_back( 0.2f );  // 2
	votes.push_back( 0.2f );  // 2
	votes.push_back( 0.25f ); // 2
	vector<float> histvalues = hist( votes, 7 );

	if ( histvalues[ histvalues.size()- 1 ] == 2.0f/11.0f && sum( histvalues ) == 1.0f )
	{
		std::cout << "Pass!" << std::endl;
		return true;
	}
	else
	{
		std::cout << "Fail." << std::endl;
		return false;
	}
}


#endif