/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#ifndef _SCVMAT_H_
#define _SCVMAT_H_

#include <cxcore.h>
#include <math.h>
#include <fstream>

template <typename t_val>
class SCVMat
{
public:
	SCVMat(){pMat = NULL;}
	SCVMat(int iRows, int iCols, int iNChan = 1){CreateMat(iRows, iCols, iNChan);}
	SCVMat(const IplImage* pImg)
	{
		pMat = NULL;
		CopyFromImg(pImg);
	}
	void CopyFromImg(const IplImage* pImg)
	{
		if (pMat)
			cvReleaseMat(&pMat);
		CreateMat(pImg->height, pImg->width, pImg->nChannels);
		for (int r=0; r<pImg->height; r++)
		for (int c=0; c<pImg->width; c++)
		for (int ch=0; ch<pImg->nChannels; ch++)
			(*this)(r,c,ch) = ((unsigned char*)(pImg->imageData+r*pImg->widthStep))
				[c*pImg->nChannels+ch];
	}
	SCVMat(const SCVMat<t_val>& iMat)
	{
		pMat = NULL;
		*this = iMat;
	}
	~SCVMat(){if (pMat) cvReleaseMat(&pMat);}
	void Load(const char* iFileName);
	void Save(const char* iFileName) const;
	void LoadBinary(std::ifstream& iFile);
	void SaveBinary(std::ofstream& iFile) const;
	CvMat* GetMat(){return pMat;}
	const CvMat* GetMat() const {return pMat;}
	int GetChan()const {return  CV_MAT_CN(pMat->type);}
	int GetRowCount() const {return pMat->rows;}
	int GetColCount() const {return pMat->cols;}
	int GetEleCount() const {return GetChan() * GetRowCount() * GetColCount();}
	t_val* GetDataPtr();
	const t_val* GetDataPtr() const;
	t_val& operator()(int r, int c, int ch = 0)
	{return GetDataPtr()[
		r*pMat->step/sizeof(t_val)+c*GetChan()+ch];}
	const t_val& operator()(int r, int c, int ch = 0) const
	{return GetDataPtr()[
		r*pMat->step/sizeof(t_val)+c*GetChan()+ch];}
	t_val GetEle(int r, int c, int ch=0) const
	{
		return GetDataPtr()[
			r*pMat->step/sizeof(t_val)+c*GetChan()+ch];
	}
	SCVMat<t_val>& operator=(const SCVMat& iMat)
	{
		if (pMat)
			cvReleaseMat(&pMat);
		if (iMat.pMat)
			pMat = cvCloneMat(iMat.pMat);
		return *this;
	}
	bool operator==(const SCVMat& iMat) const
	{
		assert(iMat.GetRowCount() == this->GetRowCount());
		assert(iMat.GetColCount() == this->GetColCount());
		assert(iMat.GetChan() == this->GetChan());
		for (int r=0; r<GetRowCount(); r++)
			for (int c=0; c<GetColCount(); c++)
				for (int ch=0; ch<GetChan(); ch++)
					if ((*this)(r,c,ch)!=iMat(r,c,ch))
						return 0;
		return 1;
	}
	SCVMat<t_val>& operator/=(const t_val val)
	{
		for (int r=0; r<GetRowCount(); r++)
			for (int c=0; c<GetColCount(); c++)
				for (int ch=0; ch<GetChan(); ch++)
					(*this)(r,c,ch) /= val;
		return *this;
	}
	SCVMat<t_val>& operator*=(const t_val val)
	{
		for (int r=0; r<GetRowCount(); r++)
			for (int c=0; c<GetColCount(); c++)
				for (int ch=0; ch<GetChan(); ch++)
					(*this)(r,c,ch) *= val;
		return *this;
	}
	template<class t_Mat>
	SCVMat<t_val>& operator+=(const t_Mat& mat)
	{
		assert(mat.GetRowCount() == GetRowCount());
		assert(mat.GetColCount() == GetColCount());
		assert(mat.GetChan() == GetChan());
		for (int r=0; r<GetRowCount(); r++)
			for (int c=0; c<GetColCount(); c++)
				for (int ch=0; ch<GetChan(); ch++)
					(*this)(r,c,ch) += mat.GetEle(r,c,ch);
		return *this;
	}
	SCVMat<t_val> operator+(const SCVMat& iMat) const
	{
		SCVMat retMat(*this);
		cvScaleAdd(this->pMat, cvScalarAll(1), iMat.pMat, retMat.pMat);
		return retMat;
	}
	SCVMat<t_val> operator-(const SCVMat& iMat) const
	{
		SCVMat retMat(*this);
		cvScaleAdd(iMat.pMat, cvScalarAll(-1), this->pMat, retMat.pMat);
		return retMat;
	}
	SCVMat<t_val> operator-() const
	{
		return -1*(*this);
	}
	SCVMat<t_val> operator*(const SCVMat& iMat) const
	{
		SCVMat retMat(GetRowCount(), iMat.GetColCount(), GetChan());
		cvGEMM(this->pMat, iMat.pMat, 1, NULL, 0, retMat.pMat);
		return retMat;
	}
	SCVMat<t_val> Transpose() const
	{
		SCVMat retMat(this->GetColCount(), this->GetRowCount(),
			this->GetChan());
		cvTranspose(this->pMat, retMat.pMat);
		return retMat;
	}
	SCVMat<t_val> Invert(int method = CV_LU) const
	{
		SCVMat retMat(this->GetColCount(), this->GetRowCount(),
			this->GetChan());
		cvInvert(this->pMat, retMat.pMat, method);
		return retMat;
	}
	//SCVMat<t_val> Eigenvalue() const
	//{
	//	SCVMat<t_val> retMat(this->GetColCount(), this->GetRowCount(),
	//		this->GetChan());
	//	cvEigenVV(this->GetMat(), retMat.GetMat());
	//}
	template <typename t_retVal>
	t_retVal LnNorm(int n) const
	{
		t_retVal retVal = 0;
		for (int r=0; r<GetRowCount(); r++)
		for (int c=0; c<GetColCount(); c++)
		for (int ch=0; ch<GetChan(); ch++)
			retVal+=pow(t_retVal(abs((*this)(r,c,ch))), n);
		return pow(retVal,t_retVal(1.0/n));
	}
	void LnNormalize(int n = 2)
	{
		t_val norm = LnNorm<t_val>(n);
		if (norm != 0)
			(*this) /= LnNorm<t_val>(n);
	}
	void Identity()
	{
		for (int r=0; r<GetRowCount(); r++)
			for (int c=0; c<GetColCount(); c++)
				for (int ch=0; ch<GetChan(); ch++)		
				(*this)(r,c,ch)=(r==c);
	}
	static SCVMat<t_val> Identity(int rowCount, int colCount, int chCount = 1)
	{
		SCVMat<t_val> retMat(rowCount, colCount, chCount);
		retMat.Identity();
		return retMat;
	}
	void SetValue(t_val val)
	{
		cvSet(pMat, cvScalarAll(val));
	}
	void SetValue(CvScalar val)
	{
		cvSet(pMat, val);
	}
	void SetZero()
	{
		cvZero(pMat);
	}
	t_val GetMaxVal()
	{
		t_val e = (*this)(0,0,0);
		for (int r=0; r<GetRowCount(); r++)
			for (int c=0; c<GetColCount(); c++)
				for (int ch=0; ch<GetChan(); ch++)
					e = max(e, (*this)(r,c,ch));
		return e;
	}
	t_val GetMinVal()
	{
		t_val e = (*this)(0,0,0);
		for (int r=0; r<GetRowCount(); r++)
			for (int c=0; c<GetColCount(); c++)
				for (int ch=0; ch<GetChan(); ch++)
					e = min(e, (*this)(r,c,ch));
		return e;
	}
	IplImage* GetImg(bool is_normalize = true, int ch = 0,
		float norm_fac = 0)
	{
		double maxval;
		double minval;
		if (is_normalize && norm_fac == 0)
		{
			maxval = is_normalize?GetMaxVal():1.0;
			minval = is_normalize?GetMinVal():0.0;
		}
		IplImage* retImg = cvCreateImage(
			cvSize(GetColCount(), GetRowCount()),
			IPL_DEPTH_8U, ch?ch:GetChan());
		int chn = min(retImg->nChannels, GetChan());
		for (int r=0; r<GetRowCount(); r++)
		for (int c=0; c<GetColCount(); c++)
		for (int chan=0; chan<chn; chan++)
		{
			/*int x = is_normalize?
				int(((*this)(r,c,chan) - minval) / (maxval - minval)*255):
			int((*this)(r,c,chan));*/
			(retImg->imageData+r*retImg->widthStep)[c*retImg->nChannels+chan] =
				/*(unsigned char)
				max(0.0,min(255.0,
				abs((*this)(r,c,chan))*256.0*3 / sqrt(double(GetColCount()*GetRowCount()))// + 127.0
				));*/
				(unsigned char)
				is_normalize?
				int
				(norm_fac==0?
				(((*this)(r,c,chan) - minval) / (maxval - minval)*255)
				:((*this)(r,c,chan))*norm_fac):
				int((*this)(r,c,chan));
		}
		return retImg;
	}
	t_val EleLen(int r, int c, int n = 2) const
	{
		t_val retVal = 0;
		for (int ch=0; ch<GetChan(); ch++)
			retVal += pow(abs((*this)(r,c,ch)),n);
		return (t_val)pow((double)retVal, 1.0/n);
	}
	SCVMat<t_val> GetAbs() const
	{
		SCVMat<t_val> retMat(*this);
		for (int r=0; r<retMat.GetRowCount(); r++)
			for (int c=0; c<retMat.GetColCount(); c++)
				for (int ch=0; ch<retMat.GetChan(); ch++)
					retMat(r,c,ch) = abs(retMat(r,c,ch));
		return retMat;
	}
private:
	CvMat* pMat;
	void CreateMat(int iRows, int iCols, int iNChan);
};

template <typename t_val, typename t_ele>
SCVMat<t_ele> operator*(const t_val& scale, const SCVMat<t_ele>& mat)
{
	SCVMat<t_ele> retMat(mat.GetRowCount(), mat.GetColCount(), mat.GetChan());
	for (int r=0; r<mat.GetRowCount(); r++)
	for (int c=0; c<mat.GetColCount(); c++)
		retMat(r,c) = mat(r,c) * scale;
	return retMat;
}

template <typename t_val, typename t_ele>
SCVMat<t_ele> operator*(const SCVMat<t_ele>& mat, const t_val& scale)
{
	return scale*mat;
}

template <typename t_val, typename t_ele>
SCVMat<t_ele> operator/(const SCVMat<t_ele>& mat, const t_val& scale)
{
	return (1/scale)*mat;
}

template <typename t_val>
void SCVMat<t_val>::Load(const char* iFileName)
{
	std::ifstream iFile(iFileName);
	int rows, cols;
	iFile>>rows>>cols;
	if (pMat)
	{
		cvReleaseMat(&pMat);
	}
	CreateMat(rows, cols, 1);
	for (int r=0; r<rows; r++)
		for (int c=0; c<cols; c++)
		{
			t_val val;
			iFile>>val;
			(*this)(r,c) = val;
			//iFile>>((*this)(r,c));
		}
}

template <typename t_val>
void SCVMat<t_val>::Save(const char* iFileName) const
{
	std::ofstream File(iFileName);
	File<<GetRowCount()<<" "<<GetColCount()<<std::endl;
	for (int r=0; r<GetRowCount(); r++)
	{
		for (int c=0; c<GetColCount(); c++)
			File<<(*this)(r,c)<<" ";
		File<<std::endl;
	}
}

template <typename t_val>
void SCVMat<t_val>::LoadBinary(std::ifstream& iFile)
{
	int rows, cols, chns;
	iFile.read((char*)&rows, 4);
	iFile.read((char*)&cols, 4);
	iFile.read((char*)&chns, 4);
	CreateMat(rows, cols, chns);
	iFile.read((char*)GetDataPtr(), rows*cols*chns*4);
}

template <typename t_val>
void SCVMat<t_val>::SaveBinary(std::ofstream& iFile) const
{
	int rows = GetRowCount();
	int cols = GetColCount();
	int chns = GetChan();
	iFile.write((char*)&rows, 4);
	iFile.write((char*)&cols, 4);
	iFile.write((char*)&chns, 4);
	iFile.write((char*)GetDataPtr(), rows*cols*chns*4);
}

inline void SCVMat<uchar>::CreateMat(int iRows, int iCols, int iNChan)
{
	if (iRows!=0 && iCols!=0)
	{
		pMat = cvCreateMat(iRows, iCols, CV_8UC(iNChan));
		SetZero();
	}
	else
		pMat = NULL;
}

inline void SCVMat<short>::CreateMat(int iRows, int iCols, int iNChan)
{
	if (iRows!=0 && iCols!=0)
	{
		pMat = cvCreateMat(iRows, iCols, CV_16SC(iNChan));
		SetZero();
	}
	else
		pMat = NULL;
}

inline void SCVMat<int>::CreateMat(int iRows, int iCols, int iNChan)
{
	if (iRows!=0 && iCols!=0)
	{
		pMat = cvCreateMat(iRows, iCols, CV_32SC(iNChan));
		SetZero();
	}
	else
		pMat = NULL;
}

inline void SCVMat<float>::CreateMat(int iRows, int iCols, int iNChan)
{
	if (iRows!=0 && iCols!=0)
	{
		pMat = cvCreateMat(iRows, iCols, CV_32FC(iNChan));
		SetZero();
	}
	else
		pMat = NULL;
}

inline void SCVMat<double>::CreateMat(int iRows, int iCols, int iNChan)
{
	if (iRows!=0 && iCols!=0)
	{
		pMat = cvCreateMat(iRows, iCols, CV_64FC(iNChan));
		SetZero();
	}
	else
		pMat = NULL;
}

inline uchar* SCVMat<uchar>::GetDataPtr(){return pMat->data.ptr;}
inline short* SCVMat<short>::GetDataPtr(){return pMat->data.s;}
inline int* SCVMat<int>::GetDataPtr(){return pMat->data.i;}
inline float* SCVMat<float>::GetDataPtr(){return pMat->data.fl;}
inline double* SCVMat<double>::GetDataPtr(){return pMat->data.db;}

inline const uchar* SCVMat<uchar>::GetDataPtr() const{return pMat->data.ptr;}
inline const short* SCVMat<short>::GetDataPtr() const{return pMat->data.s;}
inline const int* SCVMat<int>::GetDataPtr() const{return pMat->data.i;}
inline const float* SCVMat<float>::GetDataPtr() const{return pMat->data.fl;}
inline const double* SCVMat<double>::GetDataPtr() const{return pMat->data.db;}

#endif