/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once
#include <vector>
#include <algorithm>
#include <math.h>
//#include "RandomNumber.h"
#include "SSSSmallVec.h"

namespace SSSLib
{
	static const double Pi = 3.1415926535897932384626433832795;

	template <typename _TVal>
	int sign(_TVal val1)
	{
		if (val1==0)
			return 0;
		else if (val1>0)
			return 1;
		else
			return -1;
	}

	template <typename _TVal, typename _TFac>
	_TVal interpolate(_TVal val1, _TVal val2, _TFac intPosi) {return val1 + (val2-val1) * intPosi;}

	template <typename _TVal>
	_TVal Min(_TVal val1, _TVal val2){return val1<val2?val1:val2;}

	template <typename _TVal>
	_TVal Max(_TVal val1, _TVal val2){return val1>val2?val1:val2;}

	template <typename _TVal>
	_TVal clamp(_TVal val, _TVal lowBound, _TVal highBound) {return Min(Max(val, lowBound), highBound);}

	template <typename _TVal>
	bool within(_TVal val, _TVal lowBound, _TVal highBound) {return val>=lowBound && val<=highBound;}

	template <typename _TVal>
	_TVal sqr(_TVal val){return val*val;}

	inline int warp(int iVal, int iHighBound)
	{
		int retVal = iVal - iHighBound*(iVal/iHighBound);
		if (retVal<0)
			retVal += iHighBound;
		return retVal;
	}

	template <typename _TVal>
	_TVal warp(_TVal iVal, _TVal iHighBound)
	{
		return (_TVal)(iVal - iHighBound*floor(((double)iVal)/iHighBound));
	}

	template <typename _TVal>
	void getCDF(const _TVal* iPDF, int iN, _TVal* oCDF)
	{
		oCDF[0] = iPDF[0];
		for (int i=1; i<iN; i++)
			oCDF[i] = oCDF[i-1] + iPDF[i];
		_TVal vmax = oCDF[iN-1];
		if (vmax != 0)
		for (int i=0; i<iN; i++)
			oCDF[i] /= vmax;
	}

	template <typename _TVal>
	_TVal randf()
	{return ((_TVal)((((rand()&32767) << 15) + (rand()&32767)))) / 1073741824;}
	//{return ((_TVal)((((GetRand(omp_get_thread_num())&32767) << 15) + GetRand(omp_get_thread_num())) & 65535)) / 65536;}

	template <typename _TVal>
	int randCDF(const _TVal* iCDF, int iN)
	{
		using namespace std;
		_TVal r = randf<_TVal>();
		return Min(lower_bound(iCDF, iCDF+iN, r) - iCDF, iN-1);
	}

	template <typename _TVal>
	int randPDF(const _TVal* iPDF, int iN)
	{
		// a random floating point number
		using namespace std;
		vector<_TVal> cdf(iN);
		getCDF(iPDF, iN, (_TVal*)&cdf[0]);
		return randCDF((_TVal*)&cdf[0], iN);
	}

	inline void randGauss(float& oX, float& oY)
	{
		float u = randf<float>();
		float v = randf<float>();
		float c = sqrt(-2.0f*log(1.0f-u));
		oX = c*cos(2.0f*3.1415926535897932384626433832795f*v);
		oY = c*sin(2.0f*3.1415926535897932384626433832795f*v);
	}

	inline Vec2f randGauss()
	{
		Vec2f retVec;
		randGauss(retVec[0], retVec[1]);
		return retVec;
	}

	template <class typeT>
	float Length(std::vector<typeT>& iVec)
	{
		float len=0;
		for (unsigned int i=0; i<iVec.size(); i++)
			len+=iVec[i]*iVec[i];
		return sqrt(len);
	}

	template <class typeT>
	typeT Sum(std::vector<typeT>& iVec)
	{
		typeT sum=0;
		for (unsigned int i=0; i<iVec.size(); i++)
			sum+=iVec[i];
		return sum;
	}

	template <class typeT>
	float Avg(std::vector<typeT>& iVec)
	{
		return float(Sum(iVec))/float(iVec.size());
	}

	template <class typeT>
	float Diff(std::vector<typeT>& iVec)
	{
		float len=Length(iVec);
		float sum=(float)Sum(iVec);
		return len*len-sum*sum/float(iVec.size());
	}

	template <class typeT>
	float AvgDiff(std::vector<typeT>& iVec)
	{
		return Diff(iVec)/iVec.size();
	}

	template <class typeT>
	float Length(typeT* iVec,unsigned int iLen)
	{
		float len=0;
		for (unsigned int i=0; i<iLen; i++)
			len+=iVec[i]*iVec[i];
		return sqrt(len);
	}

	template <class typeT>
	typeT Sum(typeT* iVec,unsigned int iLen)
	{
		typeT sum=0;
		for (unsigned int i=0; i<iLen; i++)
			sum+=iVec[i];
		return sum;
	}

	template <class typeT>
	float Avg(typeT* iVec,unsigned int iLen)
	{
		return float(Sum(iVec,iLen))/float(iLen);
	}

	template <class typeT>
	float Diff(typeT* iVec,unsigned int iLen)
	{
		float len=Length(iVec,iLen);
		float sum=(float)Sum(iVec,iLen);
		return len*len-sum*sum/float(iLen);
	}

	template <class typeT>
	float AvgDiff(typeT* iVec,unsigned int iLen)
	{
		return Diff(iVec,iLen)/iLen;
	}
	
	template <class typeT>
	typeT Bound(typeT iData, typeT iLowBound, typeT iHighBound)
	{
		return max(min(iData, iHighBound), iLowBound);
	}

}