/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <GL/glew.h>
#include "OGLTexture.h"
#include <opencv/highgui.h>
#include <opencv/cxcore.h>
#include "OGLScene.h"
#include <iostream>
#include <vector>

using namespace SSSLib;
using namespace std;

GLuint OGLTexture2D::sCommonFBO = 0;
bool OGLTexture2D::sIsFBOInit = false;

GLuint OGLTexture2D::GetFBO()
{
	if (!sIsFBOInit)
	{
		int err = glGetError();
		if (err>0)
			std::cout<<"Drawing Error:"<<err<<std::endl;

		glGenFramebuffersEXT(1, &sCommonFBO);
		sIsFBOInit = true;
		OGLCheckError();
	}
	return sCommonFBO;
}

OGLTexture2D::OGLTexture2D(void)
: IsCreated(false)
{
}

OGLTexture2D::~OGLTexture2D(void)
{
	if (IsCreated)
	{
		glDeleteTextures(1, &ID);
	}
}

void OGLTexture2D::CreateTexture(GLsizei iWidth,
								 GLsizei iHeight,
								 GLint iChans,
								 GLenum iFormat,
								 GLenum iType,
								 void *iData )
{
	if (IsCreated)
		glDeleteTextures(1, &ID);
	glGenTextures(1, &ID);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, ID);
	mFormat = iFormat;
	mType = iType;
	mWidth = iWidth;
	mHeight = iHeight;
	glTexImage2D( GL_TEXTURE_2D, 0, iChans, iWidth, iHeight, 0, iFormat, iType, iData);
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP); 
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); 
	OGLCheckError();
	IsCreated = true;
}

void OGLTexture2D::LoadFromFile(const char* iFileName, SSSLib::Vec3uc* ipColorKey)
{
	IplImage* img = cvLoadImage(iFileName, CV_LOAD_IMAGE_COLOR);
	if (!img)
		return;
	if (!ipColorKey)
	{
		std::vector<unsigned char> dataBuffer(img->width*img->height*3);
		for (int r=0; r<img->height; r++)
			memcpy(&dataBuffer[(img->height-1-r)*img->widthStep],
				img->imageData+r*img->widthStep,
				img->widthStep);
		CreateTexture3ChUBBGR(img->width, img->height, &dataBuffer[0]);
	}
	else
	{
		std::vector<unsigned char> dataBuffer(img->width*img->height*4);
		std::vector<unsigned char>::iterator ite = dataBuffer.begin();
		for (int r=0; r<img->height; r++)
		for (int c=0; c<img->width; c++)
		{
			for (int i=0; i<3; i++)
			{
				*ite = img->imageData[(img->height-1-r)*img->widthStep+c*3+2-i];
				ite++;
			}
			if (*ipColorKey==(unsigned char*)img->imageData+(img->height-1-r)*img->widthStep+c*3)
			{
				*(ite-3) = 0;
				*(ite-2) = 0;
				*(ite-1) = 0;
				*ite=0;
			}
			else
				*ite=255;
			ite++;
		}
		CreateTexture4ChUB(img->width, img->height, &dataBuffer[0]);
	}

#ifdef _GET_MIPMAP
	glGenerateMipmap(GL_TEXTURE_2D);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR); 
#else //_GET_MIPMAP

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

#endif //_GET_MIPMAP

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP); 

	cvReleaseImage(&img);
}

void OGLTexture2D::Bind(bool iIsMultiTexture, int iTextureUnit)
{
	if (iIsMultiTexture)
		glActiveTexture(GL_TEXTURE0 + iTextureUnit);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, ID);

	OGLCheckError();
}

void OGLTexture2D::Unbind(int iTextureUnit)
{
	if (iTextureUnit!=-1)
		glActiveTexture(GL_TEXTURE0 + iTextureUnit);
	glDisable(GL_TEXTURE_2D);

	OGLCheckError();
}

void OGLTexture2D::BindFrameBuffer(GLenum iBufferName)
{
//	#define __MY_DEBUG

#ifdef __MY_DEBUG
		GLuint color_tex;
		GLuint fb;
		GLuint depth_rb;
		
		 glGenTextures(1, &color_tex);
  glBindTexture(GL_TEXTURE_2D, color_tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST); 
  //NULL means reserve texture memory, but texels are undefined
  glTexImage2D(GL_TEXTURE_2D, 0, 3, 256, 256, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
  //-------------------------
  //glGenFramebuffersEXT(1, &fb);
  //glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
  ////Attach 2D texture to this FBO
  //glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, color_tex, 0);
  ////-------------------------
  ////glGenRenderbuffersEXT(1, &depth_rb);
  ////glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, depth_rb);
  ////glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, 256, 256);
  //////-------------------------
  //////Attach depth buffer to FBO
  ////glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, depth_rb);
  ////-------------------------
  ////Does the GPU support current FBO configuration?
  //GLenum status;
  //status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
  //switch(status)
  //{
  //   case GL_FRAMEBUFFER_COMPLETE_EXT:
  //   cout<<"good";
	 //break;
  //default:
  //   cout<<status<<endl;
  //}
#endif
	//glEnable(GL_TEXTURE_2D);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, GetFBO());
	OGLCheckError();
	glFramebufferTexture2D(GL_FRAMEBUFFER_EXT,
		iBufferName,
		GL_TEXTURE_2D,	
		GetID(),
		//color_tex,
		0);

	GLenum status;
	status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
	switch(status)
	{
	case GL_FRAMEBUFFER_COMPLETE:
		break;
	default:
		cout<<"Frame Buffer Error:"<<status<<endl;
	}

	OGLCheckError();
}

void OGLTexture2D::SetDrawBuffer(int iNum, GLenum *ipNames)
{
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, GetFBO());
	OGLCheckError();
	//OGLCheckError();
	if (ipNames != NULL)
		glDrawBuffers(iNum, ipNames);
	else
	{
		vector<GLenum> names(iNum);
		for (int i=0; i<iNum; i++)
			names[i] = GL_COLOR_ATTACHMENT0_EXT+i;
		glDrawBuffers(iNum, &names[0]);
	}
	OGLCheckError();
}

void OGLTexture2D::ReadCurrentBuffer(GLenum iBufferName, GLenum iFormat, GLenum iType, int iWidth, int iHeight, GLvoid* oData)
{
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, GetFBO());
	OGLCheckError();
	glReadBuffer(iBufferName);
	OGLCheckError();
	glReadPixels(0, 0, iWidth, iHeight, iFormat, iType, oData);
}