/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "SSSVolMesh.h"
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <Eigen/Sparse>
#include <Eigen/SparseLU>
#include "SSSMeshUtil.h"
#include <algorithm>
#include "../SSSSmallMat.h"

using namespace std;
using namespace SSSLib;

void SSSVolMesh::ComputeMass(double iDensity)
{
	mSolidAngles.resize(mTets.size());
	mMass.clear();
	mMass.resize(mVtx.size(), 0);
	vector<double> weight(mVtx.size(), 0);

//#define _LOCAL_DEBUG
#ifdef _LOCAL_DEBUG
	vector<double> sumSolid(mVtx.size(), 0);
#endif

	for (int i=0; i<(int)mTets.size(); i++)
	{
		Vec4i tet = mTets[i];
		double w[4];
		double vol;
		for (int v=0; v<4; v++)
		{
			Vec3d edges[3];
			for (int j=0; j<3; j++)
			{
				edges[j] = mVtx[tet[(j+1+v)%4]]-mVtx[tet[v]];
			}
			if (v%2==1)
				swap(edges[0], edges[1]);
			if (v==0)
				vol = TripleProduct<double>(edges[0], edges[1], edges[2])/6;
			double angle = SolidAnge(edges);
			//assert(angle>1e-3);
			w[v] = angle;
			mSolidAngles[i][v] = angle;
#ifdef _LOCAL_DEBUG
			sumSolid[tet[v]] += angle;
#endif
		}
		double sumw = w[0]+w[1]+w[2]+w[3];
		for (int v=0; v<4; v++)
		{
			mMass[tet[v]] += w[v]/sumw*vol;
			weight[tet[v]] += w[v]/sumw;
		}
//		assert(fabs(w[0]+w[1]+w[2]+w[3]-4*Pi)<1e-6);
	}


	for (int i=0; i<(int)mHexs.size(); i++)
	{
		_THex hex = mHexs[i];
		Vec3d ctr(0, 0, 0);
		for (int j=0; j<8; j++)
		{
			ctr += mVtx[hex[j]];
		}
		ctr /= 8.0f;

		double vol = 0;

		Vec3d faceVtx0[4];
		Vec3d faceVtx[3];

		for (int fdir=0; fdir<3; fdir++)
		for (int fs=0; fs<2; fs++)
		for (int t=0; t<2; t++)
		{
			for (int v=0; v<4; v++)
				faceVtx0[v] = mVtx[hex[
									(fs<<fdir)
										+(((v&1)^fs)<<((fdir+1)%3))
										+((((v>>1)&1))<<((fdir+2)%3))]];
			if (t==0)
				for (int j=0; j<3; j++)
					faceVtx[j] = faceVtx0[j];
			else
				for (int j=0; j<3; j++)
					faceVtx[j] = faceVtx0[3-j];

			double incvol = dot<double>(ctr-faceVtx[0], cross3(faceVtx[1]-faceVtx[0], faceVtx[2]-faceVtx[0]))/3;
			assert(incvol>0);
			vol += incvol;
		}

		for (int j=0; j<8; j++)
			mMass[hex[j]] += vol/8.0f;
	}


	for (int v=0; v<(int)mVtx.size(); v++)
	{
		//mMass[v] = mMass[v]/weight[v]*iDensity;
		mMass[v] *= iDensity;
		//mMass[v] = 1;
	}
}

void SSSVolMesh::VtxRelax()
{
	//mSurfaceMesh.VtxPlanarRelaxProc();
	BuildVtxConnect();
	mVtx = mSurfaceMesh.mVtx;
	ZeroLapRelax(mVtxConnect, mbIntVtx, mVtx);
}

void SSSVolMesh::HexVtxRelax()
{
	mSurfaceMesh.VtxPlanarRelaxProc();
	BuildHexVtxConnect();
	//mVtx = mSurfaceMesh.mVtx;
	ZeroLapRelax(mVtxConnect, mbIntVtx, mVtx);
}

int SSSVolMesh::CheckCrossTet()
{
	int retVal = 0;

	for (int i=0; i<(int)mTets.size(); i++)
	{
		cout<<"Checking "<<i<<endl;
		Vec4i tet0 = mTets[i];
		vector<int> tet0set;
		tet0set.assign(tet0.val, tet0.val+4);
		sort(tet0set.begin(), tet0set.end());
		for (int j=i+1; j<(int)mTets.size(); j++)
		{
			Vec4i tet1 = mTets[j];
			vector<int> tet1set;
			tet1set.assign(tet1.val, tet1.val+4);
			sort(tet1set.begin(), tet1set.end());
			vector<int> cross(4);
			vector<int>::iterator ite = set_intersection(
												tet0set.begin(), tet0set.end(),
												tet1set.begin(), tet1set.end(),
												cross.begin());
			cross.resize(ite-cross.begin());

			if (cross.size()==3)
			{
				Vec3d norm = TriangleCross(mVtx[cross[0]], mVtx[cross[1]], mVtx[cross[2]]);
				vector<int> v0(1);
				vector<int> v1(1);
				set_difference(tet0set.begin(), tet0set.end(),
								cross.begin(), cross.end(),
								v0.begin());
				set_difference(tet1set.begin(), tet1set.end(),
								cross.begin(), cross.end(),
								v1.begin());

				double b = dot<double>(norm, mVtx[cross[0]]);
				if ((dot<double>(norm, mVtx[v0[0]])-b)*(dot<double>(norm, mVtx[v1[0]])-b)<0)
					retVal++;
			}
			else if (cross.size()==4)
				cout<<"Duplicated tet: ("<<i<<','<<j<<")\n";
		}
	}
	return retVal;
}

void SSSVolMesh::BuildCubeMesh(Vec3i* iCubeNodes,
							   int iNumNodes,
							   Vec3d iTranslate,
							   double iScale,
							   std::vector<Vec2i>* oEdges,
							   std::vector<Vec2i>* oHexEdges)
{
	if (oEdges)
		oEdges->clear();

	set<Vec3i> cubeNodeSet;
	map<Vec3i, int> cubeNodeID;
	for (int i=0; i<iNumNodes; i++)
	{
		cubeNodeSet.insert(iCubeNodes[i]);

		for (int x=0; x<2; x++)
		for (int y=0; y<2; y++)
		for (int z=0; z<2; z++)
		{
			Vec3i newNode = iCubeNodes[i];
			newNode[0] += x;
			newNode[1] += y;
			newNode[2] += z;

			if (cubeNodeID.count(newNode)==0)
			{
				cubeNodeID[newNode] = mVtx.size();
				mVtx.push_back(Vec3d(0.0f, 0.0f, 0.0f));
			}
		}
	}

	int tetBeg = mTets.size();
	for (int i=0; i<iNumNodes; i++)
	{
		//int newNodeList[8];

		_THex hex;

		for (int x=0; x<2; x++)
		for (int y=0; y<2; y++)
		for (int z=0; z<2; z++)
		{
			Vec3i newNode = iCubeNodes[i];
			newNode[0] += x;
			newNode[1] += y;
			newNode[2] += z;

			hex[x+(y<<1)+(z<<2)] = cubeNodeID[newNode];

			//if (cubeNodeID.count(newNode)==0)
			//{
			//	cubeNodeID[newNode] = mVtx.size();
			//	mVtx.push_back(Vec3d(0.0f, 0.0f, 0.0f));
			//}

			//newNodeList[x*4+y*2+z] = cubeNodeID[newNode];

			if (oHexEdges)
			{
				if (x==0)
					oHexEdges->push_back(Vec2i(cubeNodeID[newNode], cubeNodeID[newNode+Vec3i(1, 0, 0)]));
				if (y==0)
					oHexEdges->push_back(Vec2i(cubeNodeID[newNode], cubeNodeID[newNode+Vec3i(0, 1, 0)]));
				if (z==0)
					oHexEdges->push_back(Vec2i(cubeNodeID[newNode], cubeNodeID[newNode+Vec3i(0, 0, 1)]));
			}
		}

		mHexs.push_back(hex);

		for (int xyz=0; xyz<3; xyz++)
		for (int d=0; d<2; d++)
		{
			Vec3i newNode = iCubeNodes[i];
			Vec3i neighbour = newNode;
			neighbour[xyz] += d*2-1;
			if (cubeNodeSet.count(neighbour)==0)
			{
				Vec4i quad;
				for (int x1=0; x1<2; x1++)
				for (int x2=0; x2<2; x2++)
				{
					Vec3i node = newNode;
					node[xyz] += d;
					node[(xyz+1)%3] += x1;
					node[(xyz+2)%3] += x2;
					assert(cubeNodeID.count(node)>0);
					quad[x1+x2*2] = cubeNodeID[node];
				}
				swap(quad[2], quad[3]);
				mSurfaceMesh.mQuad.push_back(quad);
			}
		}

		vector<int> oddNodes(0);
		for (int x=0; x<2; x++)
		for (int y=0; y<2; y++)
		for (int z=0; z<2; z++)
		if ((x+y+z)%2==0)
		{
			Vec4i newTet;

			Vec3i newNode = iCubeNodes[i];
			newNode[0] += x;
			newNode[1] += y;
			newNode[2] += z;

			int xyz[3] = {x,y,z};

			newTet[0] = cubeNodeID[newNode];
			Vec3i xyznodes[3];

			for (int j=0; j<3; j++)
				xyznodes[j] = newNode;
			xyznodes[0][0] += 1-x*2;
			xyznodes[1][1] += 1-y*2;
			xyznodes[2][2] += 1-z*2;
			for (int j=0; j<3; j++)
				newTet[j+1] = cubeNodeID[xyznodes[j]];

			for (int j=0; j<3; j++)
			{
				Vec3i neighbour = iCubeNodes[i];
				neighbour[j] += xyz[j]*2-1;
				if (cubeNodeSet.count(neighbour)==0)
				{
					Vec3i face;
					face[0] = cubeNodeID[newNode];
					face[1] = cubeNodeID[xyznodes[(j+1)%3]];
					face[2] = cubeNodeID[xyznodes[(j+2)%3]];
					if (dot<int>(Vec3i(x, y, z),
								cross3(xyznodes[(j+1)%3]-newNode, xyznodes[(j+2)%3]-newNode))<0)
						swap(face[1], face[2]);
					mSurfaceMesh.mFace.push_back(face);
				}
			}

			if (oEdges)
			for (int j=0; j<3; j++)
			{
				Vec3i neighbour[2] = {iCubeNodes[i], iCubeNodes[i]};
				for (int k=0; k<2; k++)
					neighbour[k][(j+k+1)%3] += xyz[(j+k+1)%3]*2-1;

				if (!cubeNodeSet.count(neighbour[0]) && !cubeNodeSet.count(neighbour[1]))
				{
					oEdges->push_back(Vec2i(cubeNodeID[newNode], cubeNodeID[xyznodes[j]]));
				}
			}

			mTets.push_back(newTet);
		}
		else
			oddNodes.push_back(cubeNodeID[iCubeNodes[i]+Vec3i(x,y,z)]);
		mTets.push_back(Vec4i(oddNodes[0], oddNodes[1], oddNodes[2], oddNodes[3]));
	}

	for (map<Vec3i, int>::iterator ite = cubeNodeID.begin();
		ite != cubeNodeID.end();
		ite++)
	{
		Vec3i node = ite->first;
		Vec3d pst = node;
		pst += iTranslate;
		pst *= iScale;
		mVtx[ite->second] = pst;
	}

	for (int i=tetBeg; i<(int)mTets.size(); i++)
	if (TetSignedVol(i)<0)
		swap(mTets[i][2], mTets[i][3]);

	mSurfaceMesh.mVtx = mVtx;
}

void SSSVolMesh::FindIntVtx()
{
	std::vector<double> angles(mVtx.size(), 0);
	mbIntVtx.resize(mVtx.size());
	for (int i=0; i<(int)mHexs.size(); i++)
	{
		_THex hex = mHexs[i];
		for (int v=0; v<8; v++)
		{
			Vec3d vtx = mVtx[hex[v]];
			Vec3d edges[3];
			int s = 1;
			for (int e=0; e<3; e++)
			{
				int vv = v;
				int ebit = (1<<e);
				if ((vv & ebit)>0)
				{
					vv-=ebit;
					s*=-1;
				}
				else
					vv+=ebit;
				edges[e] = mVtx[hex[vv]]-vtx;
			}
			if (s<0)
				swap(edges[1], edges[2]);
			double angle = SolidAnge(edges);
			angles[hex[v]] += angle;
		}
	}

	for (int t=0; t<(int)mTets.size(); t++)
	{
		Vec4i tet = mTets[t];
		for (int v=0; v<4; v++)
		{
			Vec3d vtx = mVtx[tet[v]];
			Vec3d edges[3];
			int s = 1;
			for (int e=0; e<3; e++)
				edges[e] = mVtx[tet[(v+e+1)%4]]-vtx;

			if (v%2==1)
				swap(edges[0], edges[1]);

			double angle = SolidAnge(edges);
			angles[tet[v]] += angle;
		}
	}

	mID2SurfID.clear();
	mID2SurfID.resize(mVtx.size());
	mSurfID2ID.clear();

	for (int v=0; v<(int)mVtx.size(); v++)
	{
		//if (v==366)
		//	cout<<"!";
		if (fabs(angles[v]-4*Pi)<1e-3)
		{
			mbIntVtx[v] = true;
			mID2SurfID[v] = -1;
		}
		else
		{
			mbIntVtx[v] = false;
			mID2SurfID[v] = mSurfID2ID.size();
			mSurfID2ID.push_back(v);
		}
	}

}

void SSSVolMesh::BuildSurfaceMesh(bool ibBuildTriMesh, bool ibBuildQuadMesh)
{
	mSurfaceMesh.mVtx = mVtx;
	mSurfaceMesh.mFace.clear();
	mSurfaceMesh.mQuad.clear();

	// tet
	if (ibBuildTriMesh)
	{
		map<Vec3i, int> faceCount;
		mbSurfaceTet.clear();
		mbSurfaceTet.resize(mTets.size(), false);

		for (int i=0; i<(int)mTets.size(); i++)
		{
			Vec4i tet = mTets[i];
			for (int f=0; f<4; f++)
			{
				Vec3i face;
				for (int j=0; j<3; j++)
					face[j] = tet[(f+j)%4];
				if (f%2==0)
					swap(face[1], face[2]);
				sort(face.val, face.val+3);
				faceCount[face] = faceCount[face]+1;
			}		
		}
		for (int i=0; i<(int)mTets.size(); i++)
		{
			Vec4i tet = mTets[i];
			for (int f=0; f<4; f++)
			{
				Vec3i face;
				for (int j=0; j<3; j++)
					face[j] = tet[(f+j)%4];
				if (f%2==0)
					swap(face[1], face[2]);
				Vec3i face0 = face;
				sort(face.val, face.val+3);
				if (faceCount[face]<2)
				{
					mSurfaceMesh.mFace.push_back(face0);
					mbSurfaceTet[i] = true;
				}
			}
		}
	}
	// hex
	{

		map<Vec4i, int> faceCount;

		for (int i=0; i<(int)mHexs.size(); i++)
		{
			_THex hex = mHexs[i];
			for (int f=0; f<6; f++)
			{
				Vec4i face = GetHexFace(f);
				for (int j=0; j<4; j++)
					face[j] = hex[face[j]];
				sort(face.val, face.val+4);
				faceCount[face] = faceCount[face]+1;
			}		
		}
		for (int i=0; i<(int)mHexs.size(); i++)
		{
			_THex hex = mHexs[i];
			for (int f=0; f<6; f++)
			{
				Vec4i face = GetHexFace(f);
				for (int j=0; j<4; j++)
					face[j] = hex[face[j]];
				Vec4i face0 = face;
				sort(face.val, face.val+4);
				if (faceCount[face]<2)
				{
					if (ibBuildTriMesh)
					{
						mSurfaceMesh.mFace.push_back(Vec3i(face0[0], face0[1], face0[2]));
						mSurfaceMesh.mFace.push_back(Vec3i(face0[3], face0[2], face0[1]));
					}
					if (ibBuildQuadMesh)
					{
						Vec4i quad = face0;
						swap(quad[2], quad[3]);
						mSurfaceMesh.mQuad.push_back(quad);
					}
				}
			}
		}
	}
}

Vec4i SSSVolMesh::GetHexFace(int iFaceNum)
{
	Vec4i retVal;
	int c = iFaceNum/2;
	int s = iFaceNum%2;
	int bot = s<<c;
	for (int c1=0; c1<2; c1++)
	for (int c2=0; c2<2; c2++)
	{
		int val = bot;
		val += (c1^s)<<((c+1)%3);
		val += c2<<((c+2)%3);
		retVal[c1*2+c2] = val;
	}
	return retVal;
}

void SSSVolMesh::BuildHexVtxConnect(bool ibIsInit)
{
	if (ibIsInit)
	{
		mVtxConnect.clear();
		mVtxConnect.resize(mVtx.size());
	}
	for (int h=0; h<(int)mHexs.size(); h++)
	{
		_THex hex = mHexs[h];
		for (int v=0; v<8; v++)
		{
			int vid0 = hex[v];
			for (int c=0; c<3; c++)
			{
				int v1 = v;
				v1 = v1^(1<<c);
				mVtxConnect[vid0].insert(hex[v1]);
			}
		}
	}
}

void SSSVolMesh::ComputeHexEdges(std::vector<Vec2i>& oEdges)
{
	set<Vec2i> edges;
	for (int i=0; i<(int)mHexs.size(); i++)
	{
		_THex hex = mHexs[i];
		for (int p=0; p<7; p++)
		for (int c=0; c<3; c++)
		{
			int p2 = p;
			if ((p2&(1<<c))==0)
			{
				p2 += (1<<c);
				Vec2i edge(hex[p], hex[p2]);
				sort(edge.val, edge.val+2);
				edges.insert(edge);
			}
		}
	}
	oEdges.assign(edges.begin(), edges.end());
}

Vec3i SSSVolMesh::GetTetFace(Vec4i iTet, int iID)
{
	assert(iID<4);
	Vec3i face;
	for (int i=0; i<3; i++)
		face[i] = iTet[(iID+i)%4];
	return face;
}

Vec2i SSSVolMesh::GetTetEdge(Vec4i iTet, int iID)
{
	assert(iID<6);
	if (iID<3)
		return Vec2i(iTet[0], iTet[iID+1]);
	else if (iID<5)
		return Vec2i(iTet[1], iTet[iID-3+2]);
	else if (iID<6)
		return Vec2i(iTet[2], iTet[3]);
	else 
		return Vec2i(0, 0);
}

void SSSVolMesh::Tet2Hex()
{
	map<Vec2i, int> edges;
	map<Vec3i, int> faces;
	map<int, int> ctrs;
	for (int t=0; t<(int)mTets.size(); t++)
	{
		Vec4i tet = mTets[t];
		// enumerate faces
		for (int c=0; c<4; c++)
		{
			Vec3i face = GetTetFace(tet, c);
			sort(face.val, face.val+3);
			if (faces.count(face)==0)
			{
				faces[face] = mVtx.size();
				mVtx.push_back((mVtx[face[0]]+mVtx[face[1]]+mVtx[face[2]])/3);
			}
		}

		for (int c=0; c<6; c++)
		{
			Vec2i edge = GetTetEdge(tet, c);
			sort(edge.val, edge.val+2);
			if (edges.count(edge)==0)
			{
				edges[edge] = mVtx.size();
				mVtx.push_back((mVtx[edge[0]]+mVtx[edge[1]])/2);
			}
		}
		ctrs[t] = mVtx.size();
		mVtx.push_back((mVtx[tet[0]]+mVtx[tet[1]]+mVtx[tet[2]]+mVtx[tet[3]])/4);
	}

	// build hex
	for (int t=0; t<(int)mTets.size(); t++)
	{
		Vec4i tet = mTets[t];
		for (int i=0; i<4; i++)
		{
			int node = tet[i];
			Vec3i neighbours;
			for (int c=0; c<3; c++)
				neighbours[c] = tet[(i+c+1)%4];
			if (i%2!=0)
				swap(neighbours[0], neighbours[1]);
			
			int edge[3];
			int face[3];
			for (int j=0; j<3; j++)
			{
				Vec2i e(node, neighbours[j]);
				sort(e.val, e.val+2);
				assert(edges.count(e));
				edge[j] = edges[e];

				Vec3i f(node, neighbours[j], neighbours[(j+1)%3]);
				sort(f.val, f.val+3);
				assert(faces.count(f));
				face[j] = faces[f];
			}

			_THex hex;
			hex[0] = node;
			hex[1] = edge[0];
			hex[2] = edge[1];
			hex[3] = face[0];
			hex[4] = edge[2];
			hex[5] = face[2];
			hex[6] = face[1];
			hex[7] = ctrs[t];

			mHexs.push_back(hex);
		}
	}
}

void SSSVolMesh::Translate(Vec3d iVec)
{
	for (int i=0; i<(int)mVtx.size(); i++)
	{
		mVtx[i] = mVtx[i]+iVec;
	}

	for (int i=0; i<(int)mSurfaceMesh.mVtx.size(); i++)
	{
		mSurfaceMesh.mVtx[i] = mSurfaceMesh.mVtx[i]+iVec;
	}
}

void SSSVolMesh::Scale(double iScale)
{
	for (int i=0; i<(int)mVtx.size(); i++)
	{
		mVtx[i] *= iScale;
	}

	for (int i=0; i<(int)mSurfaceMesh.mVtx.size(); i++)
	{
		mSurfaceMesh.mVtx[i] *= iScale;
	}
}

void SSSVolMesh::Scale(Vec3d iScale)
{
	for (int i=0; i<(int)mVtx.size(); i++)
	for (int c=0; c<3; c++)
	{
		mVtx[i][c] *= iScale[c];
	}

	for (int i=0; i<(int)mSurfaceMesh.mVtx.size(); i++)
	for (int c=0; c<3; c++)
	{
		mSurfaceMesh.mVtx[i][c] *= iScale[c];
	}
}

void SSSVolMesh::Rotate(Vec3d iAxis, double iAngle)
{
	Mat3x3d rotateMat;
	double sinTheta = sin(iAngle);
	double cosTheta = cos(iAngle);
	
	for (int i=0; i<3; i++)
	{
		rotateMat(i, i) = cosTheta+sqr(iAxis[i])*(1-cosTheta);
		for (int j=1; j<3; j++)
			rotateMat(i, (i+j)%3) = iAxis[i]*iAxis[(i+j)%3]*(1-cosTheta)
									+iAxis[(i+3-j)%3]*sinTheta*(j*2-3);
	}

	for (int i=0; i<(int)mVtx.size(); i++)
	{
		mVtx[i] = rotateMat*mVtx[i];
	}

	for (int i=0; i<(int)mSurfaceMesh.mVtx.size(); i++)
	{
		mSurfaceMesh.mVtx[i] = rotateMat*mSurfaceMesh.mVtx[i];
	}
}

void SSSVolMesh::RemoveIsolatedVtx()
{
	vector<int> vtxCount(mVtx.size(), 0);

	for (int i=0; i<(int)mTets.size(); i++)
	{
		Vec4i& tet = mTets[i];
		for (int j=0; j<4; j++)
			vtxCount[tet[j]]++;
	}

	for (int i=0; i<(int)mHexs.size(); i++)
	{
		_THex& hex = mHexs[i];
		for (int j=0; j<8; j++)
			vtxCount[hex[j]]++;
	}

	vector<int> vtxNewID(mVtx.size());
	vtxNewID[0] = 0;
	for (int i=1; i<(int)mVtx.size(); i++)
	{
		if (vtxCount[i-1]==0)
			vtxNewID[i] = vtxNewID[i-1];
		else
			vtxNewID[i] = vtxNewID[i-1]+1;
	}

	for (int i=(int)mVtx.size()-1; i>=0; i--)
	if (vtxCount[i]==0)
		mVtx.erase(mVtx.begin()+i);

	for (int i=0; i<(int)mTets.size(); i++)
	{
		Vec4i& tet = mTets[i];
		for (int j=0; j<4; j++)
			tet[j] = vtxNewID[tet[j]];
	}

	for (int i=0; i<(int)mHexs.size(); i++)
	{
		_THex& hex = mHexs[i];
		for (int j=0; j<8; j++)
			hex[j] = vtxNewID[hex[j]];
	}

}

double SSSVolMesh::AvgEdgeLength()
{
	double retVal = 0;
	int edgeCount = 0;
	//for (int i=0; i<(int)mTets.size(); i++)
	//{
	//	Vec4i& tet = mTets[i];
	//	for (int v1=0; v1<4; v1++)
	//	for (int v2=v1+1; v2<4; v2++)
	//	{
	//		retVal += (mVtx[tet[v1]]-mVtx[tet[v2]]).Length();
	//		edgeCount ++;
	//	}
	//}

	assert(mVtxConnect.size()>0);

	for (int i=0; i<(int)mVtxConnect.size(); i++)
	for (set<int>::iterator ite = mVtxConnect[i].begin();
		ite != mVtxConnect[i].end();
		ite++)
	{
		retVal += (mVtx[i]-mVtx[*ite]).Length();
		edgeCount++;
	}

	return retVal/edgeCount;
}

