/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "SSSMesh.h"
#include <fstream>
#include <sstream>
#include <string>
#include "SSSMeshUtil.h"
#include "../SSSFunc.h"

namespace SSSLib
{

class SSSVolMesh
{
public:
	typedef SSSLib::SmallVec<int, 8> _THex;

	SSSVolMesh(void);
	~SSSVolMesh(void);

	void LoadVolMesh(const char* iFileName);
	void LoadNodes(const char* iFileName);
	void LoadFaces(const char* iFileName);
	void LoadTets(const char* iFileName);
	void LoadInp(const char* iFileName);
	void LoadDisteneMesh(const char* iFileName);
	void WriteNodes(const char* oFileName)
	{WriteNodes(oFileName, mVtx);}
	void WriteMesh(const char* iFileName);
	void WriteAbaqusMsh(const char* oFileName);
	bool mbLoadZeroBase;

	void AddPoissonSamples(double iPr,
		std::vector<SSSLib::Vec3d>& ioPS,
		std::vector<double>* ipSize);

	static void WriteNodes(const char* oFileName, std::vector<SSSLib::Vec3d>& iNodeList);

	template <class _TEle, int _EleSize>
	static void WriteArray(std::ofstream& oFile, std::vector<_TEle>& iVec)
	{
		for (int i=0; i<(int)iVec.size(); i++)
		{
			oFile<<i+1<<'\t';
			for (int j=0; j<3; j++)
				oFile<<iVec[i][j]<<'\t';
			oFile<<std::endl;
		}
	}

	template <class _TEle, int _EleSize>
	void LoadArray(const char* iFileName, std::vector<_TEle>& oVec)
	{
                using namespace std;
		ifstream inFile(iFileName);
		string line;
		int count;
		std::getline(inFile, line);
		{
			stringstream lineStream(line);
			lineStream>>count;
		}

		oVec.resize(count);

		for (int i=0; i<count; i++)
		{
			std::getline(inFile, line);
			stringstream lineStream(line);
			int id;
			_TEle ele;
			lineStream>>id;
			if (id==0)
				mbLoadZeroBase = true;
			if (!mbLoadZeroBase)
				id--;
			for (int j=0; j<_EleSize; j++)
				lineStream>>ele[j];
			oVec[id] = ele;
		}
	}
	
	SSSVtxSet mVtx;
	std::vector<bool> mbIntVtx;
	std::vector<int> mID2SurfID;
	std::vector<int> mSurfID2ID;
	std::vector<SSSLib::Vec4i> mTets;
	std::vector<bool> mbSurfaceTet;
	std::vector<_THex> mHexs;
	//std::vector<int> mVtxSurfaceMatch;
	std::vector<std::set<int> > mVtxConnect;
	std::vector<double> mMass;
	std::vector<Vec4d> mSolidAngles;

	static double SolidAnge(SSSLib::Vec3d* iEdges)
	{
		Vec3d* edges = iEdges;
		double el[3];
		for (int j=0; j<3; j++)
		{
			el[j] = edges[j].Length();
		}
		double det = TripleProduct<double>(edges[0], edges[1], edges[2]);
		double div = el[0]*el[1]*el[2];
		for (int j=0; j<3; j++)
			div += dot<double>(edges[(j+1)%3], edges[(j+2)%3])*el[j];
		double retval = atan2(det, div);
		if (retval<0)
			retval += (double)Pi;
		return retval*2;
	}
	void ComputeMass(double iDensity);

	void BuildVtxConnect(bool ibIsInit = true)
	{
		if (ibIsInit)
		{
			mVtxConnect.clear();
			mVtxConnect.resize(mVtx.size());
		}
		SSSLib::BuildVtxConnect<SSSLib::Vec4i, 4>((int)mVtx.size(),
								mTets,
								mVtxConnect);
	}

	void VtxRelax();
	void HexVtxRelax();

	SSSMesh mSurfaceMesh;
	double TetSignedVol(int iTet)
	{
		Vec3d edge[3];
		Vec4i tet = mTets[iTet];
		for (int i=0; i<3; i++)
			edge[i] = mVtx[tet[i+1]]-mVtx[tet[0]];

		return dot<double>(edge[0], cross3(edge[1], edge[2]));
	}

	int CheckCrossTet();
	void BuildCubeMesh(Vec3i* iCubeNodes,
						int iNumNodes,
						Vec3d iTranslate,
						double iScale,
						std::vector<Vec2i>* oEdges,
						std::vector<Vec2i>* oHexEdges);

	// the vtk format works with hex only
	void LoadVtk(const char* iFileName);
	void WriteVtk(const char* oFileName);

	void FindIntVtx();
	void BuildSurfaceMesh(bool ibBuildTriMesh = true, bool ibBuildQuadMesh = false);

	static Vec4i GetHexFace(int iFaceNum);
	void BuildHexVtxConnect(bool ibIsInit = true);
	void ComputeHexEdges(std::vector<Vec2i>& oEdges);

	Vec3i GetTetFace(Vec4i iTet, int iID);
	Vec2i GetTetEdge(Vec4i iTet, int iID);
	void Tet2Hex();

	void Translate(Vec3d iVec);
	void Scale(double iScale);
	void Scale(Vec3d iScale);
	void Rotate(Vec3d iAxis, double iAngle);

	void RemoveIsolatedVtx();

	bool CheckInside(Vec3d iPt);
	// must build connection first
	double AvgEdgeLength();
};

}