/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "SSSSimpleSpline.h"
#include <opencv\highgui.h>

using namespace SSSLib;
using namespace std;

//SSSLib::SSSSimpleSpline::SSSSimpleSpline(void)
//: mIsClosed(true)
//{
//}

void SSSSimpleSpline::ComputePT(int i,
	SSSLib::Vec2d& p0,
	SSSLib::Vec2d& p1,
	SSSLib::Vec2d& t0,
	SSSLib::Vec2d& t1)
{
	int n = mCtrPt.size();
	if (i<0)
		i=i-n*(i/n-1);
	i%=n;
	p0 = mCtrPt[i%n];
	p1 = mCtrPt[(i+1)%n];
	if (i!=0 || mIsClosed)
		t0 = 0.5f*(mCtrPt[(i+1)%n]-mCtrPt[(i+n-1)%n]);
	else
		t0 = Vec2f(0.05f, 0);
	if (i+2<(int)mCtrPt.size() || mIsClosed)
		t1 = 0.5f*(mCtrPt[(i+2)%n]-mCtrPt[i%n]);
	else
		t1 = Vec2f(0.05f, 0);
}

Vec2d SSSSimpleSpline::GetPoint(int i, double iS)
{
	SSSLib::Vec2d p0, p1, t0, t1;
	ComputePT(i, p0, p1, t0, t1);

	double s1 = iS;
	double s2 = iS*iS;
	double s3 = s2*iS;

	double h1 =  2*s3 - 3*s2 + 1;          // calculate basis function 1
	double h2 = -2*s3 + 3*s2;              // calculate basis function 2
	double h3 =   s3 - 2*s2 + s1;         // calculate basis function 3
	double h4 =   s3 -  s2;              // calculate basis function 4
	return h1*p0+h2*p1+h3*t0+h4*t1;
}

Vec2d SSSSimpleSpline::GetGradient(double iS)
{
	SSSLib::Vec2d p0, p1, t0, t1;
	ComputePT((int)floor(iS), p0, p1, t0, t1);

	double frac = iS-floor(iS);

	double s1 = frac;
	double s2 = s1*frac;
	
	double h1 =  6*s2 - 6*s1;          // calculate basis function 1
	double h2 = -6*s2 + 6*s1;              // calculate basis function 2
	double h3 =  3*s2 - 4*s1 + 1;         // calculate basis function 3
	double h4 =  3*s2 - 2*s1;              // calculate basis function 4
	return h1*p0+h2*p1+h3*t0+h4*t1;
}

Vec2d SSSSimpleSpline::GetNormal(double iS)
{
	Vec2d grad = GetGradient(iS);
	return Vec2d(-grad[1], grad[0]);
}

vector<double> SSSSimpleSpline::ApproxLength(int iSegs)
{
	vector<double> retVec(iSegs);
	Vec2d lasPt = GetPoint(0);
	double delta = double(mCtrPt.size())/iSegs;
	for (int i=0; i<iSegs; i++)
	{
		double s = delta*(i+1);
		Vec2d newPt = GetPoint(s);
		retVec[i] = (newPt-lasPt).Length();
		if (i!=0)
			retVec[i] += retVec[i-1];
		lasPt = newPt;
	}
	return retVec;
}

cv::Mat SSSSimpleSpline::Draw(int iReso)
{
	cv::Mat outMat(iReso, iReso, CV_8UC3);
	for (int i=0;  i<(int)mCtrPt.size()-1; i++)
	{
		double step = sqrt((mCtrPt[i+1]-mCtrPt[i]).SqrLength())/iReso;
		Vec2f prevPt = mCtrPt[i];
		for (double state = step; state<1.0f; state += step)
		{
			Vec2f curPt = GetPoint(i, state);
			cv::line(outMat,
				cv::Point(int(prevPt[0]*iReso), int(prevPt[1]*iReso)),
				cv::Point(int(curPt[0]*iReso), int(curPt[1]*iReso)),
				cv::Scalar(255.0f, 255.0f, 255.0f));
			prevPt = curPt;
		}
		Vec2f curPt = mCtrPt[i+1];
		cv::line(outMat,
			cv::Point(int(prevPt[0]*iReso), int(prevPt[1]*iReso)),
			cv::Point(int(curPt[0]*iReso), int(curPt[1]*iReso)),
			cv::Scalar(255.0f, 255.0f, 255.0f));
	}
	return outMat;
}

void SSSSimpleSpline::InsertPt(SSSLib::Vec2d iPt)
{
	vector<Vec2d>::iterator ite = mCtrPt.begin();
	while (ite->val[0]<iPt[0])
		ite++;
	mCtrPt.insert(ite, iPt);
}

std::vector<SSSLib::Vec2d> SSSSimpleSpline::BuildLine(double iBegS, double iEndS, int iReso)
{
	vector<SSSLib::Vec2d> retVec(iReso+1);
	double coord = iBegS;
	double coordDelta = iEndS - iBegS;
	coordDelta /= iReso;
	for (int j=0; j<=iReso; j++)
	{
		retVec[j] = GetPoint(coord);
		coord += coordDelta;
	}
	return retVec;
}
