/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "SSSSmallVec.h"
#include <iostream>

namespace SSSLib
{
	template <typename _TVal, int _NRows, int _NCols>
	struct SmallMat;

	typedef SmallMat<double, 2, 2> Mat2x2d;
	typedef SmallMat<float, 2, 2> Mat2x2f;
	typedef SmallMat<double, 3, 3> Mat3x3d;
	typedef SmallMat<float, 3, 3> Mat3x3f;
	typedef SmallMat<double, 4, 4> Mat4x4d;
	typedef SmallMat<float, 4, 4> Mat4x4f;

	template <typename _TVal, int _NRows, int _NCols>
	struct SmallMat
	{
		int rows() const{return _NRows;}
		int cols() const{return _NCols;}
		SmallMat(){}
		SmallMat(const SmallMat<_TVal, _NRows, _NCols>& iMat)
			: data(iMat.data)
		{}
		SmallMat(const _TVal* iData)
		{
			for (int r=0; r<_NRows; r++)
				data[r] = SmallVec<_TVal, _NCols>(iData+r*_NCols);
		}
		SmallMat(
			const _TVal& a00, const _TVal& a01, const _TVal& a02, const _TVal& a03,
			const _TVal& a10, const _TVal& a11, const _TVal& a12, const _TVal& a13,
			const _TVal& a20, const _TVal& a21, const _TVal& a22, const _TVal& a23,
			const _TVal& a30, const _TVal& a31, const _TVal& a32, const _TVal& a33)
		{
			data[0][0] = a00;
			data[0][1] = a01;
			data[0][2] = a02;
			data[0][3] = a03;

			data[1][0] = a10;
			data[1][1] = a11;
			data[1][2] = a12;
			data[1][3] = a13;

			data[2][0] = a20;
			data[2][1] = a21;
			data[2][2] = a22;
			data[2][3] = a23;

			data[3][0] = a30;
			data[3][1] = a31;
			data[3][2] = a32;
			data[3][3] = a33;
		}
		static SmallMat<_TVal, _NRows, _NCols> Identity()
		{
			SmallMat<_TVal, _NRows, _NCols> retMat;
			for (int r=0; r<_NRows; r++)
			for (int c=0; c<_NCols; c++)
			if (r==c)
				retMat(r, c) = (_TVal)1;
			else
				retMat(r, c) = 0;
			return retMat;
		}
		static SmallMat<_TVal, _NRows, _NCols> Zeros()
		{
			SmallMat<_TVal, _NRows, _NCols> retMat;
			for (int r=0; r<_NRows; r++)
			for (int c=0; c<_NCols; c++)
				retMat(r, c) = 0;
			return retMat;
		}
		static SmallMat<_TVal, _NRows, _NCols> Ones()
		{
			SmallMat<_TVal, _NRows, _NCols> retMat;
			for (int r=0; r<_NRows; r++)
			for (int c=0; c<_NCols; c++)
				retMat(r, c) = 1;
			return retMat;
		}
		SmallVec<SmallVec<_TVal, _NCols>, _NRows> data;
		_TVal& operator()(int iRow, int iCol) {return data[iRow][iCol];}
		bool operator<(const SmallMat<_TVal, _NRows, _NCols>& iMat) const{return data<iMat.data;}
		_TVal* GetDataPtr(){return (_TVal*)&data;}
		const _TVal& operator()(int iRow, int iCol) const
		{return data[iRow][iCol];}
		template <int _NCols2>
		SmallMat<_TVal, _NRows, _NCols2> operator*(const SmallMat<_TVal, _NCols, _NCols2>& iMat)
		{
			SmallMat<_TVal, _NRows, _NCols2> retVal;
			for (int r=0; r<_NRows; r++)
			for (int c=0; c<_NCols2; c++)
			{
				_TVal val = 0;
				for (int k=0; k<_NCols; k++)
					val += data[r][k]*iMat(k, c);
				retVal(r, c) = val;
			}
			return retVal;
		}
		SmallMat<_TVal, _NRows, _NCols> operator+(const SmallMat<_TVal, _NRows, _NCols>& iMat)
		{
			SmallMat<_TVal, _NRows, _NCols> retVal;
			for (int r=0; r<_NRows; r++)
				retVal.data[r] = data[r]+iMat.data[r];
			return retVal;
		}
		SmallMat<_TVal, _NRows, _NCols> operator-(const SmallMat<_TVal, _NRows, _NCols>& iMat)
		{
			SmallMat<_TVal, _NRows, _NCols> retVal;
			for (int r=0; r<_NRows; r++)
				retVal.data[r] = data[r]-iMat.data[r];
			return retVal;
		}
		SmallMat<_TVal, _NRows, _NCols> operator*(const _TVal iVal) const
		{
			SmallMat<_TVal, _NRows, _NCols> retMat;
			for (int r=0; r<_NRows; r++)
				retMat.data[r] = data[r]*iVal;
			return retMat;
		}
		SmallVec<_TVal, _NRows> operator*(const SmallVec<_TVal, _NCols>& iVec) const
		{
			SmallVec<_TVal, _NRows> oVec;
			for (int i=0; i<_NRows; i++)
			{
				oVec[i] = dot<_TVal>(iVec, data[i]);
			}
			return oVec;
		}
		void Print() const
		{
			using namespace std;
			for (int r=0; r<_NRows; r++)
			{
				cout<<data[r][0];
				for (int c=1; c<_NCols; c++)
					cout<<' '<<data[r][c];
				cout<<endl;
			}
		}
		SmallMat<_TVal, _NCols, _NRows> Transpose() const
		{
			SmallMat<_TVal, _NCols, _NRows> retMat;
			for (int r=0; r<_NRows; r++)
			for (int c=0; c<_NCols; c++)
				retMat(c, r) = (*this)(r, c);
			return retMat;
		}
	};

}