/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "SSSSmallVec.h"
#include "SSSFunc.h"
#include <opencv\cxcore.h>

namespace SSSLib
{
	enum BoarderFlag
	{
		CLAMP, WARP
	};

	template <typename _TVec>
	bool within(const cv::Mat &iImg, _TVec iVec)
	{
		return within(iVec[0], 0, iImg.cols-1) && within(iVec[1], 0, iImg.rows-1);
	}

	template <typename _TMatEle, typename _TVec>
	const _TMatEle& Get(const cv::Mat &iImg, _TVec iVec)
	{
		return iImg.at<_TMatEle>(iVec[1], iVec[0]);
	}

	template <typename _TMatEle, typename _TVec>
	_TMatEle& Get(cv::Mat &iImg, _TVec iVec)
	{
		return iImg.at<_TMatEle>(iVec[1], iVec[0]);
	}

	template <typename _TMatEle>
	_TMatEle Get(const cv::Mat &iImg, float x, float y, BoarderFlag iFlag = CLAMP)
	{
		int floorx = int(floor(x));
		int floory = int(floor(y));
		int x0, x1, y0, y1;
		if (iFlag == CLAMP)
		{
			x0 = clamp(floorx, 0, iImg.cols-1);
			y0 = clamp(floory, 0, iImg.rows-1);
			x1 = clamp(floorx+1, 0, iImg.cols-1);
			y1 = clamp(floory+1, 0, iImg.rows-1);
		}
		else if (iFlag == WARP)
		{
			x0 = warp(floorx, iImg.cols);
			y0 = warp(floory, iImg.rows);
			x1 = warp(floorx+1, iImg.cols);
			y1 = warp(floory+1, iImg.rows);
		}
		return interpolate( interpolate(
			Get<_TMatEle>(iImg, Vec2i(x0, y0)),
			Get<_TMatEle>(iImg, Vec2i(x1, y0)),
			x-floorx),
			interpolate(
			Get<_TMatEle>(iImg, Vec2i(x0, y1)),
			Get<_TMatEle>(iImg, Vec2i(x1, y1)),
			x-floorx),
			y-floory);
	}

	template <typename _TIMatEle, typename _TOMatEle>
	cv::Mat Derivative(const cv::Mat& iMat)
	{
		cv::Mat retMat(iMat.size(), CV_MAKETYPE(iMat.type(),2));
		for (int r=0; r<iMat.rows; r++)
		for (int c=0; c<iMat.cols; c++)
			retMat.at<_TOMatEle>(r, c) = 0.5f*_TOMatEle(
				(iMat.at<float>(r, warp(c+1, iMat.cols))
					-iMat.at<float>(r, warp(c-1, iMat.cols)))
					,
				(iMat.at<float>(warp(r+1, iMat.rows), c)
					-iMat.at<float>(warp(r-1, iMat.rows), c))
					);
		return retMat;
	}

	template <typename _TIMatEle, typename _TOMatEle>
	cv::Mat Derivative_TopLeft(const cv::Mat& iMat)
	{
		cv::Mat retMat(iMat.size(), CV_MAKETYPE(iMat.type(),2));
		for (int r=0; r<iMat.rows; r++)
		for (int c=0; c<iMat.cols; c++)
			retMat.at<_TOMatEle>(r, c) = _TOMatEle(
				(iMat.at<float>(r, c)
					-iMat.at<float>(r, warp(c-1, iMat.cols)))
					,
				(iMat.at<float>(r, c)
					-iMat.at<float>(warp(r-1, iMat.rows), c))
					);
		return retMat;
	}

	template <typename _TIMatEle>
	_TIMatEle MatL2(const cv::Mat& iMat)
	{
		_TIMatEle retVal = 0;
		for (int r=0; r<iMat.rows; r++)
		for (int c=0; c<iMat.cols; c++)
			retVal += sqr(iMat.at<_TIMatEle>(r, c));
		return sqrt(retVal/iMat.rows/iMat.cols);
	}
}