/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "../StaticFEM/Config.h"
#include "../SSSLib/SSSLibOGL/SSSVolMesh.h"
#include "EigenUtil.h"
#include <Eigen/Sparse>
#include <set>
#include <map>
//#include "engine.h"
#include "KKTEle.h"
#include "../TinyThread++-1.1/source/tinythread.h"
//#include "../matrix-factor-master/source/solver.h"

//#include <lbfgs.h>

#define _IND_SYM_SOLVE

//#define _USE_NEG_VOL
#define _USE_SIM_LAP
//#define _EXT_SIM_LAP_ONLY
//#define _CURVATURE_WEIGHT
//#define _USE_SIM_LAP_AS_KKT

//#define _KKT_ELIMINATION

//#define _OUTPUT_LOG

//#define _2ND_BOUND
//#define _ElementSize 8

// phase method to find feasible point
// reference: section 11.4, page 579, convex optimization
//
// problem formulation:
//		minize			s
//		subject to		f_i(x)<=s
//						Ax=b
// (s: auxiliary variable)
// terminate when s<=0
//
// according to the interior point method, reformulate the problem as:
//		minimize		s - 1/t * \sum log(s-f_i(x))
//		subject to		Ax=b
// reference: section 11.2, page 562, convex optimization
// t is a parameter. We can play around with it.

class LaplacianDeformation;

class PhysImpSolve3D
{
public:
	// configure
	PhysImpSolve3D(void);
	~PhysImpSolve3D(void);

	//solver<_T_FLOAT> mSIMSolver;
#ifdef _IND_SYM_SOLVE
	void SIMSolve(Eigen::SparseMatrix<_T_FLOAT>& iA,
		const MatXFloat& ib,
		MatXFloat& ox);
#endif

	int mRecalLapStep;

	bool mbLoadState;
	bool mbNegGD;

	void ParseConfigFile(const char* iFileName,
						SSSLib::SSSVolMesh& oVolMesh);
	std::map<int, _T_FLOAT> mFixedForce;
	std::map<int, _T_FLOAT> mFixedForceDim[3];
	std::set<int> mIgnoreLap;
	std::set<int> mIgnoreLapInput;

	std::set<int> mFixedVtx;
	std::vector<SVec3Float> mVtxPresetForce;
	_T_FLOAT mJacPercentBound;
	std::vector<_T_FLOAT> mJacRef;
	std::vector<SSSLib::SmallVec<_T_FLOAT, 8> > mJacRefHex;

	bool mbMassLapWeight;
	bool mbInvLapWeight;
	_T_FLOAT mMaxDim;

	_T_FLOAT mSpeedWeight;
	void AddSpeedTerm();

	// laplacian similarity
	bool mbUseEleSim;
	_T_FLOAT mLapSimWeight;
	_T_FLOAT mLapScaleWeight;
	_T_FLOAT mLapFixedWeight;
	std::map<SSSLib::Vec2i, _T_FLOAT> mLapSimMat;
	Eigen::SparseMatrix<_T_FLOAT> mLapSimMatEigen;
	Eigen::SparseMatrix<_T_FLOAT> mLapHalfSimMatEigen;
	MatXFloat mLapSimB;
	
	void ComputeVtxWeightVM(std::vector<_T_FLOAT>& oVtxWeight);
	void ComputeVtxWeightVMLocalMax(std::vector<_T_FLOAT>& oVtxWeight);
	void ComputeVtxWeightInverseVM(std::vector<_T_FLOAT>& oVtxWeight);
	void ComputeVtxWeightNormalDiff(std::vector<_T_FLOAT>& oVtxWeight);
	void ComputeVtxWeightRelativeMass(std::vector<_T_FLOAT>& oVtxWeight);
	void ComputeVtxWeightZeroFreePointLap(std::vector<_T_FLOAT>& oVtxWeight);
	void ComputeVtxWeightLapDiff(std::vector<_T_FLOAT>& oVtxWeight,
		LaplacianDeformation& ioLap);
	void ComputeVtxWeightInverseLap(std::vector<_T_FLOAT>& oVtxWeight,
		LaplacianDeformation& ioLap);
	void EncodeVtxWeight(const std::vector<_T_FLOAT>& iVtxWeight,
						LaplacianDeformation& oLap);
	void ComputeLapAll();
	void ComputeLapMat(int iDeform, _T_FLOAT iWeight);
	void ComputeElementSimMat(int iDeform);
	void ComputeOneRingSimMat(int iDeform);
	void DecodeHalfLapMat(std::vector<Eigen::Triplet<_T_FLOAT> >& iList);
	void AddSimMatElement(int iDeform,
		_T_FLOAT iWeight,
		std::vector<int>& iElements,
		std::vector<Eigen::Triplet<_T_FLOAT> >& oList);
	int mLapSimNumRing;
	std::set<int> lNeighbour;
	void ComputeNRingNeighbour(int iNRing, LaplacianDeformation& oLap);
	_T_FLOAT EvaluateLapSim();
	void LapAddToGradient(_T_FLOAT iScale, _T_FLOAT* oGrad);
	void LapAddToKKT(_T_FLOAT iScale);

	_T_FLOAT mShapeSimWeight;

	// backup

	struct BKStates
	{
		std::vector<SVec3Float> mX0Backup;
		std::vector<SVec3Float> mP0Backup;
		std::vector<_T_FLOAT> mVMBackup;
		std::vector<SSSLib::SmallVec<_T_FLOAT, 8> > mVMHexBackup;
		std::vector<SVec3Float> mXVBackup;
		std::vector<SVec3Float> mPVBackup;
		_T_FLOAT mLastFuncVal;
		_T_FLOAT mVMBound;
		_T_FLOAT mVMWeight;
	};

	BKStates mStepBK;
	BKStates mBreakPoint;
	int mNoiseSteps;
	int mNoiseConvSteps;
	_T_FLOAT mNoiseProportion;
	_T_FLOAT mNoiseScale;

	bool NoiseBranch();
	void NoiseMerge();

	void Backup(BKStates& oBackup)
	{
		//std::cout<<"Backing Up\n";
		oBackup.mX0Backup = mX0;
		oBackup.mP0Backup = mP0;
		oBackup.mVMBackup = mVM;
		oBackup.mVMHexBackup = mVMHex;
//		mXVBackup = mXV;
//		mPVBackup = mPV;
		oBackup.mLastFuncVal = mFuncVal;

	}
	void Rollback(const BKStates& iBackup)
	{
		mX0 = iBackup.mX0Backup;
		mP0 = iBackup.mP0Backup;
		mVM = iBackup.mVMBackup;
		mVMHex = iBackup.mVMHexBackup;
//		mXV = mXVBackup;
//		mPV = mPVBackup;
		mFuncVal = iBackup.mLastFuncVal;
	}


	std::vector<SVec3Float> mXAdv;
	std::vector<SVec3Float> mPAdv;

	bool lbNoExtraVM;

	//_T_FLOAT mLastFuncVal;
	bool mbBackupInit;

	// external solver

	// hessian for laplacian solver
	std::vector<_T_FLOAT> mLapSolLambda;
	std::map<SSSLib::Vec2i, _T_FLOAT> mLapSolHessian;
	int lLapSolHessianCounter;
	int lLapSolContNum;
	bool lNeedUpdate;
	bool mbUsePenalty;
	bool mbAugmentedLag;
	_T_FLOAT mMu;
	_T_FLOAT mFuncVal;
	_T_FLOAT mFuncDropSpeed;
	std::vector<_T_FLOAT> mGrad;
	std::map<SSSLib::Vec2i, _T_FLOAT> mHess;
	_T_FLOAT mForcePenWeight;
	_T_FLOAT mFWeight;
	bool lOutputConst;
	bool lComputeHess;
	std::vector<_T_FLOAT> lConst;

	bool lbPhysValid;

	// log barrier newton solver
	bool mbComputeLapSolHessian;
	int mPhase;
	_T_FLOAT mS;
	_T_FLOAT mVS;
	_T_FLOAT mdS;
	_T_FLOAT mNewtonT;
	_T_FLOAT mNewtonTInitScale;
	_T_FLOAT mNewtonMu;
	_T_FLOAT mNewtonTCrit;
	_T_FLOAT mNewtonStep;
	_T_FLOAT mNewtonSScale;
	_T_FLOAT lNewtonEnergy;

	static bool lFirstOrderOnly;

	// two level KKT

	bool mbKKTNormStep;
	_T_FLOAT CalScoreWithPen();
	_T_FLOAT AdvanceAndEvaluate(_T_FLOAT iStep);
	bool LineSearch(_T_FLOAT iRefT, _T_FLOAT iMinT, _T_FLOAT iMaxT, _T_FLOAT& oTotalT); // assume we back-up our results already
	bool SolveEqKKT(_T_FLOAT iStep, _T_FLOAT& iXPDiagWeight, _T_FLOAT iLambdaRegular, bool ibSkipHess);
	bool EqKKTAdvance(_T_FLOAT iStep);
	void RunPhysics();
	void NormalizeKKTStep(bool ibCap);
	bool mbCholeskyAddedMultiple;

	void AddInequalityConstraint(_T_FLOAT iF,
								int iN,
								int* iIdx,
								int* ibOptimize,
								const _T_FLOAT* iGrad,
								const MatXFloat& iHess);

	SSSLib::SSSVolMesh* mpVolMesh;
	void Init(SSSLib::SSSVolMesh* ipVolMesh);

	static void Det2ndOrder(Mat3x3 iMat,
							Mat3x3& odM,
							MatXFloat& od2M);
	static MatXFloat Mat3DInvD(Mat3x3 iMat,
								Mat3x3* oMatInv = NULL,
								std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> >* od2InvdMat2 = NULL);
	static void Deform2ndOrder(const Mat3x3& iX,
								const Mat3x3& iP,
								MatXFloat& oD,
								MatXFloat& odD,
								std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> >& od2D //X then P
								);
	static void Strain2ndOrder(const Mat3x3& iX,
								const Mat3x3& iP,
								MatXFloat& oD, // [xx, yy, zz, yz, zx, xy]
								MatXFloat& odD,
								std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> >& od2D //X then P
								);
	static bool U2ndOrder(const Vec3Float* iX,
								const Vec3Float* iP,
								const MatXFloat& iE, // material tensor
								bool ibNoV,
								_T_FLOAT& oU,
								MatXFloat& odU,
								MatXFloat& od2U);

	static void Volume2ndOrder(const Vec3Float* iX,
								const Vec3Float* iP,
								_T_FLOAT& oU,
								MatXFloat& odU,
								MatXFloat& od2U);

	static _T_FLOAT ComputeU(const Vec3Float* iX,
								const Vec3Float* iP,
								const MatXFloat& iE,
								bool ibNoV,
								bool* obNegV = NULL);

	static _T_FLOAT ComputeU(const MatXFloat& iX,
								const MatXFloat& iP,
								const MatXFloat& iE,
								bool ibNoV,
								bool* obNegV = NULL);

	static MatXFloat HexBuildH(SVec3Float iCoord, int iEleSize);

	template <int _EleSize>
	static void Hex2ndOrderDecode(const MatXFloat& idU,
								const MatXFloat& id2U,
								const MatXFloat& iH,
								MatXFloat& odU,
								MatXFloat& od2U)
	{
		assert(_EleSize==iH.rows());
		const int poffset = _EleSize*3;
		odU = MatXFloat::Zero(poffset*2, 1);
		for (int xp=0; xp<2; xp++)
		{
			Mat3x3 dUMat = Vec2Mat3x3(idU.block(xp*9, 0, 9, 1)); // transposed by default
			MatXFloat odUMat = iH*dUMat; // transposed
			for (int i=0; i<_EleSize; i++)
				odU.block(xp*poffset+i*3, 0, 3, 1) = odUMat.row(i).transpose();
		}
		MatXFloat h2 = MatXFloat::Zero(_EleSize*2, 6);
		h2.block(0, 0, _EleSize, 3) = iH;
		h2.block(_EleSize, 3, _EleSize, 3) = iH;
		MatXFloat h2t(h2);
		h2t.transposeInPlace();

		if (lFirstOrderOnly)
			return;

		od2U = MatXFloat::Zero(poffset*2, poffset*2);
	
		for (int c1=0; c1<3; c1++)
		for (int c2=0; c2<3; c2++)
			Eigen::Map<Eigen::Matrix<_T_FLOAT, 2*_EleSize, 2*_EleSize>, 0, Eigen::Stride<3*poffset*2, 3> >(od2U.data()+c1+c2*od2U.rows())
				= h2*
					Eigen::Map<const Eigen::Matrix<_T_FLOAT, 2*3, 2*3>, 0, Eigen::Stride<3*3*6, 3> >(id2U.data()+c1+c2*id2U.rows())
					*h2t;
	}

	static void HexEncode(const Vec3Float* iX,
							const MatXFloat& iH,
							Vec3Float* oX);

	static void HexEncode(const Vec3Float* iX,
							const Vec3Float* iP,
							const MatXFloat& iH,
							Vec3Float* oX,
							Vec3Float* oP);

	std::vector<SVec3Float> mXRef;
	std::vector<SVec3Float> mX0;
	std::vector<SVec3Float> mP0;
	std::vector<SVec3Float> mPV;
	std::vector<SVec3Float> mXV;

	std::vector<SVec4Float> mTetVtxWeight;
	std::vector<std::set<int> > mVtxSparseStruct;
	int mNumVtxSparse;

	void BuildVtxSparseStruct();

	MatXFloat mE;
	_T_FLOAT mYoungs;
	_T_FLOAT mPR;
	static bool mbCorotation;

	std::vector<Eigen::Triplet<_T_FLOAT> > mdExtForce;

	_T_FLOAT mFDamping;
	_T_FLOAT mVDamping;
	bool mbStatic;
	_T_FLOAT mSimTimeStep;
	_T_FLOAT mXUniformMass;
	_T_FLOAT mXMassScale;
	_T_FLOAT mPMassScale;
	_T_FLOAT mDensity;
	std::vector<_T_FLOAT> mMass;
	std::vector<SVec3Float> mExtForceP;
	std::vector<SVec3Float> mExtForceX;
	_T_FLOAT mVolumeWeight;
	_T_FLOAT mForceWeight;

	std::vector<KKTEle> mKKT;

	_T_FLOAT XPStep(_T_FLOAT iDT, bool ibPOnly, bool ibKKT = true);
	void ComputeExtForce();
	void ComputePhysForce();
	void ComputePhysicsSimForce();
	void ComputeElasticForce();
	void AddDiagonalPHess(_T_FLOAT iUniformMass);
	void SetMaterialParam(_T_FLOAT iYoungs, _T_FLOAT iPoissionRatio);
	static void SolveLinear(const std::vector<KKTEle>& iSystem,
							_T_FLOAT iDiagnalWeight,
							int iIdxOffset,
							_T_FLOAT& oPenalty,
							std::vector<_T_FLOAT>& oGrad);
	static void KKTSolve(const Eigen::SparseMatrix<_T_FLOAT>& iH,
						const Eigen::SparseMatrix<_T_FLOAT>& iA,
						const MatXFloat& ig,
						const MatXFloat& ih,
						MatXFloat& ov,
						MatXFloat& ow);
	
	// P force
	bool mbGravity;
	bool mbFloor;
	_T_FLOAT mFloorHeight;
	_T_FLOAT mFloorForceWeight;

	void ComputePGravity();
	void ComputePFloor();
	void ComputePresetForce();
	void ComputeFixedForce();

	// X force
	
	// hex volume
	bool mbMinVol;
	_T_FLOAT mHexVolumeWeight;
	std::vector<_T_FLOAT> mHexDefaultVolume;
	std::vector<_T_FLOAT> mTetDefaultVolume;
	_T_FLOAT ComputeHexVolumeForce(bool ibRetTmpVal);

	// VM
	_T_FLOAT mVMBound;
	_T_FLOAT mVMWeight;
	_T_FLOAT mVMWeightMax;
	std::vector<_T_FLOAT> mVM;
	std::vector<SSSLib::SmallVec<_T_FLOAT, 8> > mVMHex;
	std::vector<_T_FLOAT> mJac;
	std::vector<SSSLib::SmallVec<_T_FLOAT, 8> > mJacHex;
	std::vector<_T_FLOAT> mVMVtx;

	enum ConttForce
	{ContVM, ContVolume, InitJac};

	void ComputeVMForce(bool ibComputeUOnly, ConttForce iCont);
	//void ComputeVolumeConst();
	void ComputeVMPerVtx();
	_T_FLOAT ComputeMaxVM() const;
	void UpdateVMBound(_T_FLOAT iNewVMBound, _T_FLOAT iS, bool ibUpdateNewtonT = false, bool ibNoIS = false);

	// Laplacian
	enum VtxType {VTInt = 0, VTSurf, VTEdge, VTPt};

	std::vector<std::set<int> > mConnect;
	std::vector<SVec3Float> mLap;
	std::vector<VtxType> mIntVtx; // internal vertices
	_T_FLOAT mLapWeight;
	_T_FLOAT mEdgeLengthEpsilon;
	std::vector<SSSLib::Vec2i>* mpEdges;
	std::vector<SSSLib::Vec2i>* mpHexEdges;
	std::vector<SSSLib::Vec2i>* mpHexExtEdges;

	void InitIntLap();
	void BuildConnect(bool ibIsQuad);
	template <int _NEleSize, class _TEle>
	void BuildConnect(VtxType iLevel, const std::vector<_TEle>& iEle)
	{
		for (int f=0; f<(int)iEle.size(); f++)
		{
			_TEle tet = iEle[f];
			for (int i=0; i<_NEleSize; i++)
			for (int j=i+1; j<_NEleSize; j++)
			if (mIntVtx[tet[i]] == iLevel && mIntVtx[tet[j]] == iLevel)
			{
				mConnect[tet[i]].insert(tet[j]);
				mConnect[tet[j]].insert(tet[i]);
			}
		}

		for (int v=0; v<(int)mX0.size(); v++)
		if (mIntVtx[v] == iLevel)
		{
			mLap[v] = ComputeLaplacianWithNeighbour(v, mpVolMesh->mVtx, mConnect[v]);
			if (mLap[v].Length()>mEdgeLengthEpsilon//*mConnect[v].size()
				|| mConnect[v].size()==0)
			{
				mIntVtx[v] = (VtxType)(mIntVtx[v]+1);
				mConnect[v].clear();
			}
		}
	}
	//void BuildLap();
	void ComputeLapForce();

	template <class _TVec>
	SVec3Float ComputeLaplacianWithNeighbour(int iVtxID, _TVec& iVtx, std::set<int>& iNeighbour)
	{
                using namespace std;
		SVec3Float retVec(0, 0, 0);
		std::set<int>& connect = iNeighbour;
		for (set<int>::iterator v1=connect.begin();
			v1!=connect.end(); v1++)
			retVec += iVtx[*v1];
		if (connect.size()>0)
			retVec = SVec3Float(iVtx[iVtxID])-retVec/connect.size();
		return retVec;
	}

	void ComputeFixedDesignForce();
	std::set<int> mFixedDesignConst;
	std::set<int> mFixedDesignDimConst[3];
	std::set<int> mFixedDesignAny;
 
	//symmetry
	_T_FLOAT mSymWeight[2];
	static SVec3Float Transform(SVec3Float iVec, MatXFloat iMat);
	void ComputeModelSymForce();
	struct TSymConst // mTransform*mV1=mV2
	{
		int mV1;
		int mV2;
		SMat4x4 mTransform;
	};
	std::vector<TSymConst> mSymConst[2];
	void AddSymConstraints(int iV1, int iV2, MatXFloat iTransform, bool ibAutoInv, bool ibPhysics);
	void AddRotateCenterConstraint(int iV1, int iV2);
	void AddMirrorConstraint(int iV1, int iV2);

	std::vector<KKTEle> mIneqJacob;
	std::vector<KKTEle> mEqJacob;

	void OptExportLinear();
	void OptExport();
	void OptExportPen();

	/*static lbfgsfloatval_t OptExportLBFGSCallback(
										void *instance,
										const lbfgsfloatval_t *iX,
										lbfgsfloatval_t *oG,
										const int iN,
										const lbfgsfloatval_t iStep);

	static int OptExportLBFGSProgress(
				void *instance,
				const lbfgsfloatval_t *x,
				const lbfgsfloatval_t *g,
				const lbfgsfloatval_t fx,
				const lbfgsfloatval_t xnorm,
				const lbfgsfloatval_t gnorm,
				const lbfgsfloatval_t step,
				int n,
				int k,
				int ls);*/
	static void LBFGSOptThread(void* argv);
	
	MatXFloat lXGuess;
	void OptSolveEqKKT();
	void OptExportPenInit();
	void OptExportMatlabInit();
	/*
	void OptExportMatlab();
	void OptExportMatlabPen();
	void OptExportMatlabSnopt();
	void MatlabSimpleTest();
	static void OptExportMatlabCallbackFPen(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[]);
	static void OptExportMatlabCallbackF(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[]);
	static void OptExportMatlabCallbackCon(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[]);
	static void OptExportMatlabCallbackHess(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[]);

	static void OptExportMatlabCallbackHead(mxArray* iX);
	static void OptExportMatlabCallbackCOnly(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[]);
	static void OptExportMatlabCallbackdC(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[]);
	static void OptExportMatlabCallbackFOnly(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[]);
	static void OptExportMatlabCallbackdF(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[]);
	static void OptExportMatlabCallbackd2F(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[]);

	mxArray* OptExportMatlabComputeHessPattern();
	mxArray* OptExportMatlabComputeConsPattern();

	mxArray* EncodeState();
	bool CopybackState(const double* iVec);
	bool CopybackState(mxArray* iXP);
	void CopyLapLambda(mxArray* iLambda);
	*/

	static PhysImpSolve3D* spPhysImp;

	//Engine* mMatlabEngine;

	static void MatlabOptThread(void* argv);
	void RunPhysics(double iConv, int iIteNum = -1);
	//static void (*RenderCallback)();
	tthread::thread* mpOptThread;

	void DumpState(const char* oFileName = "tmp.data");
	void LoadState(const char* iFileName = "tmp.data");
};

