/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PhysImpSolve3D.h"
#include "EigenUtil.h"

using namespace std;
using namespace SSSLib;

//#define _NON_V 2e-4
//#define _NON_V 1
//#define _OUTPUT_LOG

_T_FLOAT PhysImpSolve3D::ComputeU(const Vec3Float* iX,
								const Vec3Float* iP,
								const MatXFloat& iE,
								bool ibNoV,
								bool* obNegV)
{
	Mat3x3 X;
	Mat3x3 P;
	for (int i=0; i<3; i++)
	{
		X.col(i) = iX[i];
		P.col(i) = iP[i];
	}
	return ComputeU(X, P, iE, ibNoV, obNegV);
}

_T_FLOAT PhysImpSolve3D::ComputeU(const MatXFloat& iX,
								const MatXFloat& iP,
								const MatXFloat& iE,
								bool ibNoV,
								bool* obNegV)
{
	Mat3x3 d = iP*iX.inverse();
	if (mbCorotation)
	{
		Eigen::JacobiSVD<Mat3x3> svd(d, Eigen::ComputeFullU | Eigen::ComputeFullV);
		d = Mat3x3::Zero();
		VectorXFloat lambda = svd.singularValues();
		
		Mat3x3 matU = svd.matrixU();
		Mat3x3 matV = svd.matrixV();
		if (matU.determinant()<0)
			lambda(2) *= -1;
		if (matV.determinant()<0)
			lambda(2) *= -1;
		for (int i=0; i<3; i++)
			d(i, i) = lambda(i);
	}
	d = d-Mat3x3::Identity();
	d=(d+Mat3x3(d).transpose())/2;
	MatXFloat vecd(6,1);
	if (obNegV)
		*obNegV = false;
	for (int i=0; i<3; i++)
	{
		vecd(i, 0)= d(i, i);
		vecd(i+3, 0) = d((i+1)%3, (i+2)%3);
	}
#ifdef _NON_V
	return _NON_V*(MatXFloat(vecd).transpose()*iE*vecd)(0, 0);
#else
	if (ibNoV)
		return (MatXFloat(vecd).transpose()*iE*vecd)(0, 0);
	else
	{
		_T_FLOAT xdet = iX.determinant();
		if (xdet<=0 && obNegV)
			*obNegV = true;
		return xdet*(MatXFloat(vecd).transpose()*iE*vecd)(0, 0);
	}
#endif
}

void PhysImpSolve3D::Det2ndOrder(Mat3x3 iMat,
							Mat3x3& odM,
							MatXFloat& od2M)
{
	Vec3Float v[3];
	for (int i=0; i<3; i++)
		v[i] = iMat.col(i);

	// dM
	for (int i=0; i<3; i++)
		odM.col(i) = v[(i+1)%3].cross(v[(i+2)%3]);

	// for debug
	//assert((odM-iMat.determinant()*iMat.transpose().inverse()).norm()<iMat.norm()*1e-6);
	//{
	//	MatXFloat dMtmp = iMat.determinant()*iMat.inverse();
	//	cout<<"dDet a:\n"<<odM<<endl
	//		<<"dDet b:\n"<<dMtmp<<endl;
	//}

	// d2M
	if (lFirstOrderOnly)
		return;
	od2M = MatXFloat::Zero(9, 9);
	for (int r1=0; r1<3; r1++)
	for (int c1=0; c1<3; c1++)
	for (int rr=1; rr<3; rr++)
	for (int cc=1; cc<3; cc++)
	{
		int r2 = (r1+rr)%3;
		int c2 = (c1+cc)%3;
		int r3 = 3-r1-r2;
		int c3 = 3-c1-c2;
		int sign = 1;
		
		if (r1>r2)
			sign = -sign;
		if (r1>r3)
			sign = -sign;
		if (r2>r3)
			sign = -sign;

		if (c1>c2)
			sign = -sign;
		if (c1>c3)
			sign = -sign;
		if (c2>c3)
			sign = -sign;

		_T_FLOAT val1 = sign*iMat(r3, c3);
		od2M(r1*3+c1, r2*3+c2) = val1;

		// for debug
		//{
		//	Mat3x3 m1 = Mat3x3::Zero();
		//	Mat3x3 m2 = Mat3x3::Zero();
		//	m1(r1, c1) = 1;
		//	m2(r2, c2) = 1;

		//	Mat3x3 minv = iMat.inverse();
		//	_T_FLOAT val = iMat.determinant()*((minv*m1).trace()*(minv*m2).trace()-(minv*m1*minv*m2).trace());
		//	assert(fabs(val-od2M(r1*3+c1, r2*3+c2))<iMat.norm()*1e-6);
		//}
	}

}

MatXFloat PhysImpSolve3D::Mat3DInvD(Mat3x3 iMat,
									Mat3x3* oMatInv,
									std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> >* od2InvdMat2)
{
	Mat3x3 matInv = iMat.inverse();
	MatXFloat retMat(9, 9);
	for (int r=0; r<3; r++)
	for (int c=0; c<3; c++)
	{
		//Eigen::Matrix<_T_FLOAT, -1, -1, Eigen::RowMajor> tmpMat = -matInv.col(r)*matInv.row(c);
		//tmpMat.conservativeResize(9, 1);
		//retMat.col(r*3+c) = tmpMat;

		Mat3x3 tmpMat = -matInv.col(r)*matInv.row(c);
		retMat.col(r*3+c) = Mat3x3ToVec(tmpMat);
	}

	if (oMatInv)
		*oMatInv = matInv;

	if (od2InvdMat2 && !lFirstOrderOnly)
	{
		od2InvdMat2->clear();
		od2InvdMat2->resize(9, MatXFloat::Zero(9, 9));

		for (int i=0; i<9; i++)
		for (int j=0; j<9; j++)
		{
			int r1 = i/3;
			int c1 = i%3;
			int r2 = j/3;
			int c2 = j%3;

			Mat3x3 tmpMat = matInv.col(r1)*matInv(c1,r2)*matInv.row(c2)
							+matInv.col(r2)*matInv(c2,r1)*matInv.row(c1);
			for (int k=0; k<9; k++)
				(*od2InvdMat2)[k](i, j) = tmpMat(k/3, k%3);
		}
	}
	return retMat;
}

void PhysImpSolve3D::Deform2ndOrder(const Mat3x3& iX,
								const Mat3x3& iP,
								MatXFloat& oD,
								MatXFloat& odD,
								std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> >& od2D)
{
	Mat3x3 X = iX;
	Mat3x3 P = iP;

	Mat3x3 XInv;
	MatXFloat dXInv;
	std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> > d2XInv;

	dXInv = Mat3DInvD(X, &XInv, &d2XInv);

	// oD
	oD = P*XInv-Mat3x3::Identity(3, 3);
#ifdef _OUTPUT_LOG
	cout<<"D\n"<<oD<<endl;
#endif

	// dDdP
	MatXFloat dDdP = MatXFloat::Zero(9, 9);
	for (int r=0; r<3; r++)
	for (int c=0; c<3; c++)
	{
		Vec3Float data = XInv.row(c);
		dDdP.block(r*3, r*3+c, 3, 1) = data;
	}
#ifdef _OUTPUT_LOG
	cout<<"dDdP\n"<<dDdP<<endl;
#endif
	// dDdX
	MatXFloat dDdX = MatXFloat::Zero(9, 9);
	for (int i=0; i<9; i++)
	{
		//Eigen::Matrix<_T_FLOAT, -1, -1, Eigen::RowMajor> tmpMat = dXInv.col(i);
		//tmpMat.conservativeResize(3, 3);
		//tmpMat = P*tmpMat;
		//tmpMat.conservativeResize(9, 1);
		//dDdX.col(i) = tmpMat;

		Mat3x3 tmpMat = Vec2Mat3x3(dXInv.col(i));
		Mat3x3 tmpMat2 = P*tmpMat;
		dDdX.col(i) = Mat3x3ToVec(tmpMat2);
	}
#ifdef _OUTPUT_LOG
	cout<<"dDdX\n"<<dDdX<<endl;
#endif
	odD = MatXFloat(9, 18);
	odD<<dDdX, dDdP;

	if (lFirstOrderOnly)
		return;

	// d2D
	od2D.clear();
	od2D.resize(9, MatXFloat::Zero(18, 18));
	// d2DdPdX
	for (int p=0; p<9; p++)
	for (int x=0; x<9; x++)
	{
		Mat3x3 dp = Mat3x3::Zero();
		dp(p/3, p%3) = 1;
		//Eigen::Matrix<_T_FLOAT, -1, -1, Eigen::RowMajor> dx = dXInv.col(x);
		//dx.conservativeResize(3, 3);

		Mat3x3 dx = Vec2Mat3x3(dXInv.col(x));

		MatXFloat d = dp*dx;
		for (int i=0; i<9; i++)
		{
			od2D[i](p+9, x) = d(i/3, i%3);
			od2D[i](x, p+9) = d(i/3, i%3);
		}
	}
	// d2Dd2X
	for (int x0=0; x0<9; x0++)
	for (int x1=0; x1<9; x1++)
	{
		Mat3x3 d2x;
		for (int i=0; i<9; i++)
			d2x(i/3, i%3) = d2XInv[i](x0, x1);
		Mat3x3 tmpMat = P*d2x;
		for (int i=0; i<9; i++)
			od2D[i](x0, x1) = tmpMat(i/3, i%3);
	}
}

void PhysImpSolve3D::Strain2ndOrder(const Mat3x3& iX,
								const Mat3x3& iP,
								MatXFloat& oD,
								MatXFloat& odD,
								std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> >& od2D)
{
	MatXFloat D;
	MatXFloat dD;
	vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> > d2D;

	Deform2ndOrder(iX, iP, D, dD, d2D);
	//cout<<D<<endl;
	//cout<<dD<<endl;
	D = (D+MatXFloat(D).transpose())/2;
#ifdef _OUTPUT_LOG
	cout<<"d+d^t\n"<<D<<endl;
#endif
	oD = MatXFloat(6, 1);
	for (int i=0; i<3; i++)
	{
		oD(i, 0) = D(i, i);
		oD(i+3, 0) = D((i+1)%3, (i+2)%3);
	}
	
	MatXFloat tmpdD = dD;
	for (int r=0; r<3; r++)
	for (int c=0; c<3; c++)
		dD.row(r*3+c) += tmpdD.row(c*3+r);
	odD = MatXFloat(6, 18);
	for (int i=0; i<3; i++)
	{
		odD.row(i) = 0.5*dD.row(i*3+i);
		odD.row(i+3) = 0.5*dD.row(((i+1)%3)*3+(i+2)%3);
	}

	if (lFirstOrderOnly)
		return;

	std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> > tmpd2D = d2D;
	for (int r=0; r<3; r++)
	for (int c=0; c<3; c++)
		d2D[r*3+c] += tmpd2D[c*3+r];

	od2D.clear();
	od2D.resize(6, MatXFloat::Zero(18, 18));
	for (int i=0; i<3; i++)
	{
		od2D[i] = 0.5*d2D[i*3+i];
		od2D[i+3] = 0.5*d2D[((i+1)%3)*3+(i+2)%3];
		//assert(d2D[((i+1)%3)*3+(i+2)%3]==d2D[((i+2)%3)*3+(i+1)%3]);
	}
}

bool PhysImpSolve3D::U2ndOrder(const Vec3Float* iX,
								const Vec3Float* iP,
								const MatXFloat& iE, // material tensor
								bool ibNoV,
								_T_FLOAT& oU,
								MatXFloat& odU,
								MatXFloat& od2U)
{
	bool retVal = true;
	// for debug
#ifdef _NON_V
	ibNoV = true;
#endif

	Mat3x3 X;
	Mat3x3 P;
	for (int i=0; i<3; i++)
	{
		X.col(i) = iX[i];
		P.col(i) = iP[i];
	}

	Mat3x3 matU;
	Mat3x3 matV;
	
	if (mbCorotation)
	{
		Mat3x3 matF = P*X.inverse();
		Eigen::JacobiSVD<Mat3x3> svd(matF, Eigen::ComputeFullU | Eigen::ComputeFullV);
		matU = svd.matrixU();
		matV = svd.matrixV();
		//assert(matU.determinant()>=0);
		//assert(matV.determinant()>=0);
		if (matU.determinant()<0)
			matU.col(2) *= -1;
		if (matV.determinant()<0)
			matV.col(2) *= -1;
		X = matV.transpose()*X;
		P = matU.transpose()*P;
	}
	else
	{
		matU = Mat3x3::Identity();
		matV = Mat3x3::Identity();
	}

	//assert(X.determinant()>0);
#ifdef _OUTPUT_LOG
	cout<<"X "<<endl
		<<X<<endl
		<<"P \n"
		<<P<<endl;
#endif
	MatXFloat D;
	MatXFloat dD;
	vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> > d2D;
	//Strain2ndOrder(iX, iP, D, dD, d2D);
	Strain2ndOrder(X, P, D, dD, d2D);
#ifdef _OUTPUT_LOG
	cout<<"D\n"<<D<<endl;
	cout<<"dD\n"<<dD<<endl;
#endif
	
	_T_FLOAT V;
	Mat3x3 dVdX;
	MatXFloat d2VdX2;
	MatXFloat dV;
	MatXFloat dVdXVec;

	if (ibNoV)
	{
#ifdef _NON_V
		V = _NON_V;
#else
		V = 1;
#endif
		dVdX = Mat3x3::Zero();
		dVdXVec = MatXFloat::Zero(9, 1);
		d2VdX2 = MatXFloat::Zero(9, 9);
		dV = MatXFloat::Zero(18, 1);
	}
	else
	{
		V = X.determinant()/6;
		//assert(V>0);
		Det2ndOrder(X, dVdX, d2VdX2);
		dVdX /= 6;
		d2VdX2 /= 6;
		dVdXVec = Mat3x3ToVec(dVdX);
		//dVdXVec = MatXFloat(9, 1);
		//for (int i=0; i<3; i++)
		//	dVdXVec.block(i*3, 0, 3, 1) = dVdX.col(i);
		dV = MatXFloat::Zero(18, 1);
		dV.block(0, 0, 9, 1) = dVdXVec;
	}

#ifndef _USE_NEG_VOL
	if (V<=0)
		retVal = false;
#endif

	MatXFloat Dt(D);
	Dt.transposeInPlace();
	MatXFloat VDtE = V*Dt*iE;
	_T_FLOAT ded = (Dt*iE*D)(0, 0);
	oU = V*ded;
	odU = (2*VDtE*dD).transpose();
	odU.block(0, 0, 9, 1) += ded*dVdXVec;

	// transpose the output
	MatXFloat tmpdU = odU;
	for (int r=0; r<3; r++)
	{
		for (int c=0; c<3; c++)
		{
			odU.row(r*3+c) = tmpdU.row(c*3+r);
			odU.row(r*3+c+9) = tmpdU.row(c*3+r+9);
		}
		if (mbCorotation)
		{
			MatXFloat tmp = odU.block(r*3, 0, 3, 1);
			odU.block(r*3, 0, 3, 1) = matV*tmp;
			tmp = odU.block(r*3+9, 0, 3, 1);
			odU.block(r*3+9, 0, 3, 1) = matU*tmp;
		}
	}

	if (lFirstOrderOnly)
		return retVal;
	// od2U
	{
		MatXFloat tmp = 2*dV*Dt*iE*dD;

		od2U = 2*V*MatXFloat(dD).transpose()*iE*dD
			+ tmp + MatXFloat(tmp).transpose();
	}

	for (int i=0; i<18; i++)
	for (int j=0; j<18; j++)
	{
		MatXFloat tmp2(6, 1);
		for (int k=0; k<6; k++)
			tmp2(k) = d2D[k](i, j);
		od2U(i, j) += 2*(VDtE*tmp2)(0, 0);

		if (i<9 && j<9)
			od2U(i, j) += d2VdX2(i, j)*ded;
	}

	MatXFloat tmpd2U(18, 18);
	for (int r=0; r<3; r++)
	for (int c=0; c<3; c++)
	{
		tmpd2U.row(r*3+c) = od2U.row(c*3+r);
		tmpd2U.row(r*3+c+9) = od2U.row(c*3+r+9);
	}
	for (int r=0; r<3; r++)
	for (int c=0; c<3; c++)
	{
		od2U.col(r*3+c) = tmpd2U.col(c*3+r);
		od2U.col(r*3+c+9) = tmpd2U.col(c*3+r+9);
	}

	if (mbCorotation)
	{
		for (int r=0; r<3; r++)
		{
			MatXFloat tmp = od2U.block(r*3, 0, 3, od2U.cols());
			od2U.block(r*3, 0, 3, od2U.cols()) = matV*tmp;
			tmp = od2U.block(r*3+9, 0, 3, od2U.cols());
			od2U.block(r*3+9, 0, 3, od2U.cols()) = matU*tmp;
		}

		for (int c=0; c<3; c++)
		{
			MatXFloat tmp = od2U.block(0, c*3, od2U.rows(), 3);
			od2U.block(0, c*3, od2U.rows(), 3) = tmp*matV.transpose();
			tmp = od2U.block(0, c*3+9, od2U.rows(), 3);
			od2U.block(0, c*3+9, od2U.rows(), 3) = tmp*matU.transpose();

		}
	}

#ifdef _OUTPUT_LOG
	cout<<"gradient\n"
		<<odU<<endl;
	cout<<"od2U\n"
		<<od2U<<endl;
#endif
	return retVal;
}

void PhysImpSolve3D::Volume2ndOrder(const Vec3Float* iX,
								const Vec3Float* iP,
								_T_FLOAT& oU,
								MatXFloat& odU,
								MatXFloat& od2U)
{
	Mat3x3 X;
	Mat3x3 P;
	for (int i=0; i<3; i++)
	{
		X.col(i) = iX[i];
		P.col(i) = iX[i];
	}

	Mat3x3 dM;
	MatXFloat d2M;

	Det2ndOrder(X, dM, d2M);
	oU = -X.determinant();
	odU = MatXFloat::Zero(18, 1);
	od2U = MatXFloat::Zero(18, 18);
	for (int r=0; r<3; r++)
	for (int c=0; c<3; c++)
		odU(r+c*3, 0) = -dM(r, c);

	if (lFirstOrderOnly)
		return;

	for (int r1=0; r1<3; r1++)
	for (int c1=0; c1<3; c1++)
	for (int r2=0; r2<3; r2++)
	for (int c2=0; c2<3; c2++)
		od2U(r1+c1*3, r2+c2*3) = -d2M(r1*3+c1, r2*3+c2);
}
