/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PhysImpSolve3D.h"
#include "../LaplacianDeformation/LaplacianDeformation.h"
#include "PhysImpSolve3D_util.h"
#include <math.h>

using namespace std;
using namespace SSSLib;
using namespace PhysImpUtil;

void PhysImpSolve3D::ComputeLapAll()
{
	if (mbUseEleSim)
		//ComputeElementSimMat(0);
		ComputeOneRingSimMat(0);
	else
	{
		mLapSimMatEigen = Eigen::SparseMatrix<_T_FLOAT>(0, 0);
		mLapSimMat.clear();
		mLapSimB = MatXFloat::Zero(0, 0);
		//if (mLapSimWeight!=0)
			ComputeLapMat(0, mLapSimWeight);
		if (mLapScaleWeight!=0)
			ComputeLapMat(LaplacianDeformation::DEFORM_SCALE, mLapScaleWeight);
		if (mLapFixedWeight!=0)
			ComputeLapMat(LaplacianDeformation::DEFORM_NON, mLapFixedWeight);
	}
}

void PhysImpSolve3D::ComputeVtxWeightVM(std::vector<_T_FLOAT>& oVtxWeight)
{
	oVtxWeight.clear();
	oVtxWeight.resize(mXRef.size(), 0);

	ComputeVMForce(true, ContVM);

	vector<_T_FLOAT> weightw(mXRef.size(), 0);
	for (int h=0; h<mpVolMesh->mHexs.size(); h++)
	{
		SSSVolMesh::_THex& hex = mpVolMesh->mHexs[h];
		SmallVec<_T_FLOAT, 8>& vm = mVMHex[h];
		for (int i=0; i<8; i++)
		{
			int j = hex[i];
			weightw[j] += 1;
			oVtxWeight[j] += vm[i];
		}
	}

	for (int t=0; t<mpVolMesh->mTets.size(); t++)
	{
		Vec4i tet = mpVolMesh->mTets[t];
		_T_FLOAT vm = mVM[t];
		for (int i=0; i<4; i++)
		{
			int j = tet[i];
			weightw[j] += 1;
			oVtxWeight[j] += vm;
		}
	}

	for (int i=0; i<(int)oVtxWeight.size(); i++)
	if (weightw[i]>0)
	{
		oVtxWeight[i] /= weightw[i];
	}
}

void PhysImpSolve3D::ComputeVtxWeightVMLocalMax(std::vector<_T_FLOAT>& oVtxWeight)
{
	std::vector<_T_FLOAT> vm;
	ComputeVtxWeightVM(vm);

	oVtxWeight.clear();
	oVtxWeight.resize(mXRef.size(), 1e-5);

	for (int v=0; v<(int)mXRef.size(); v++)
	{
		set<int>& list_j =  mpVolMesh->mVtxConnect[v];
		for (set<int>::iterator ite = list_j.begin();
			ite != list_j.end();
			ite++)
		if (vm[v]<vm[*ite])
		{
			oVtxWeight[v] = 1;
			break;
		}
	}
}

void PhysImpSolve3D::ComputeVtxWeightInverseVM(std::vector<_T_FLOAT>& oVtxWeight)
{
	std::vector<_T_FLOAT> vm;
	ComputeVtxWeightVM(vm);

	oVtxWeight.clear();
	oVtxWeight.resize(mXRef.size(), 1e-5);

	for (int i=0; i<(int)mXRef.size(); i++)
	{
		oVtxWeight[i] = mVMBound/(mVMBound*1e-1+vm[i])*5;
		oVtxWeight[i] = Max(oVtxWeight[i], 1e-5);
	}
}

void PhysImpSolve3D::ComputeVtxWeightNormalDiff(std::vector<_T_FLOAT>& oVtxWeight)
{
	std::vector<vector<SVec3Float> > incidentNormals(mXRef.size());
	std::vector<vector<_T_FLOAT> > angleWeights(mXRef.size());

	assert(mpVolMesh->mSurfaceMesh.mQuad.size()>0);

	for (int q=0; q<(int)mpVolMesh->mSurfaceMesh.mQuad.size(); q++)
	{
		Vec4i quad = mpVolMesh->mSurfaceMesh.mQuad[q];

		for (int i=0; i<4; i++)
		{
			int vtx[3] = {quad[i], quad[(i+3)%4], quad[(i+1)%4]};
			SVec3Float edge[2];

			for (int j=0; j<2; j++)
				edge[j] = mXRef[vtx[j+1]]-mXRef[vtx[0]];

			SVec3Float normal = cross3(edge[1], edge[0]);
			_T_FLOAT cosine = dot<_T_FLOAT>(edge[1], edge[0]);
			_T_FLOAT angle = acos(cosine);
			incidentNormals[vtx[0]].push_back(normal.Norm());
			angleWeights[vtx[0]].push_back(angle);
		}
	}

	oVtxWeight.resize(mXRef.size());

	for (int i=0; i<(int)mXRef.size(); i++)
	if (incidentNormals[i].size()>0)
	{
		Vec3f vtxNormal(0, 0, 0);
		for (int j=0; j<(int)incidentNormals[i].size(); j++)
			vtxNormal += incidentNormals[i][j]*angleWeights[i][j];

		vtxNormal = vtxNormal.Norm();
		_T_FLOAT diff = 0;

		for (int j=0; j<(int)incidentNormals[i].size(); j++)
			diff += (vtxNormal-incidentNormals[i][j]).Length()*angleWeights[i][j]/2/M_PI;

		oVtxWeight[i] = exp(-diff*2+0.5);
		//oVtxWeight[i] = 0.5/(diff+1e-3);
		//oVtxWeight[i] = 1;
	}
}

void PhysImpSolve3D::ComputeVtxWeightRelativeMass(std::vector<_T_FLOAT>& oVtxWeight)
{
	oVtxWeight.resize(mXRef.size());
	_T_FLOAT avgMass = 0;
	for (int i=0; i<(int)mMass.size(); i++)
		avgMass += mMass[i];
	avgMass /= mMass.size();
	for (int i=0; i<(int)mMass.size(); i++)
		oVtxWeight[i] *= mMass[i]/avgMass;
}

void PhysImpSolve3D::ComputeVtxWeightZeroFreePointLap(std::vector<_T_FLOAT>& oVtxWeight)
{
	oVtxWeight.resize(mXRef.size());

	for (int i=0; i<(int)mXRef.size(); i++)
	if (mIntVtx[i]==VTPt)
		oVtxWeight[i] = 0;
	else
		oVtxWeight[i] = 1;
}

void PhysImpSolve3D::ComputeVtxWeightLapDiff(std::vector<_T_FLOAT>& oVtxWeight,
											 LaplacianDeformation& ioLap)
{
	oVtxWeight.resize(ioLap.point_list.size());
	ioLap.mLapWeight.resize(ioLap.point_list.size());
	for (int i=0; i<(int)ioLap.vertex_laplacian_vector.size(); i++)
	{
		vector<int>& neighbour = ioLap.laplacian_matrix[i];
		_T_FLOAT minLap = -1;
		for (int j=0; j<(int)neighbour.size(); j++)
		{
			_T_FLOAT newDis = trimesh::len(ioLap.vertex_laplacian_vector[i]
								-ioLap.vertex_laplacian_vector[neighbour[j]]);
			if (minLap<0 || newDis<minLap)
				minLap = newDis;
		}

		ioLap.mLapWeight[i] = exp(-minLap*10);
		//ioLap.mLapWeight[i] = exp(-minLap*2);
		//if (minLap<1e-1)
		//	ioLap.mLapWeight[i] = 1;
		//else
		//	ioLap.mLapWeight[i] = 1e-5;
		if (mFixedDesignConst.count(i))
			ioLap.mLapWeight[i] = 0;

		oVtxWeight[i] = ioLap.mLapWeight[i];
	}
}

void PhysImpSolve3D::ComputeVtxWeightInverseLap(std::vector<_T_FLOAT>& oVtxWeight,
											 LaplacianDeformation& ioLap)
{
	oVtxWeight.resize(ioLap.point_list.size());
	ioLap.mLapWeight.resize(ioLap.point_list.size());
	_T_FLOAT avgLap = 0;
	for (int i=0; i<(int)ioLap.vertex_laplacian_vector.size(); i++)
		avgLap += trimesh::len(ioLap.vertex_laplacian_vector[i]);
	avgLap /= ioLap.mTransformNeighbourVtx.size();
	for (int i=0; i<(int)ioLap.vertex_laplacian_vector.size(); i++)
	{
		//vector<int>& neighbour = ioLap.laplacian_matrix[i];
		//_T_FLOAT minLap = -1;
		//for (int j=0; j<(int)neighbour.size(); j++)
		//{
		//	_T_FLOAT newDis = trimesh::len(ioLap.vertex_laplacian_vector[i]
		//						-ioLap.vertex_laplacian_vector[neighbour[j]]);
		//	if (minLap<0 || newDis<minLap)
		//		minLap = newDis;
		//}

		double len  = trimesh::len(ioLap.vertex_laplacian_vector[i]);
		len /= avgLap;
		ioLap.mLapWeight[i] = 1/(len+1e-2)*oVtxWeight[i];
		if (mFixedDesignConst.count(i))
			ioLap.mLapWeight[i] = 0;

		oVtxWeight[i] = ioLap.mLapWeight[i];
	}
}

void PhysImpSolve3D::EncodeVtxWeight(const std::vector<_T_FLOAT>& iVtxWeight,
									LaplacianDeformation& oLap)
{
	oLap.mLapWeight.resize(oLap.point_list.size());
	for (int i=0; i<(int)oLap.point_list.size(); i++)
	{
		int idx;
#ifdef _EXT_SIM_LAP_ONLY
		idx = mpVolMesh->mSurfID2ID[i];
#else
		idx = i;
		
#endif
		oLap.mLapWeight[i] = iVtxWeight[idx];
		if (mFixedDesignConst.count(idx))
			oLap.mLapWeight[i] = 0;
	}
}

void PhysImpSolve3D::ComputeNRingNeighbour(int iNRing, LaplacianDeformation& oLap)
{
	lNeighbour.clear();
	oLap.laplacian_matrix.clear();
	//lap.laplacian_matrix.resize(mXRef.size());

	for(size_t i = 0; i < mXRef.size(); i++)
#ifdef _EXT_SIM_LAP_ONLY
	//if (!mpVolMesh->mbIntVtx[i])
	if (mIntVtx[i] != VTInt)
#endif
	{
		// for debug
		//if (i==2 || i==49 || i==117)
		//	cout<<i;

		vector<int> is_visited;
		is_visited.resize(mXRef.size(), 0);
		is_visited[i] = 1;
		vector<int> near_list;
		near_list.push_back(i);
		if (mIgnoreLapInput.count(i))
			mIgnoreLap.insert(i);
		int begin_ind = 0;
		int end_ind = near_list.size();
		for(int level = 0; level < iNRing; level++)
		{

			if(level != 0)
			{
				begin_ind = end_ind;
				end_ind = near_list.size();
			}
			for(size_t j = begin_ind; j < end_ind; j++)
			{
				set<int>& list_j =  mpVolMesh->mVtxConnect[near_list[j]];

				for (set<int>::iterator ite = list_j.begin();
					ite != list_j.end();
					ite++)
				if (*ite!=i)
				{
					int near_id = *ite;
#ifdef _EXT_SIM_LAP_ONLY
					if (mIntVtx[near_id]!=VTInt)
#endif
					// surface vtx only plays with surface vtx, but internal vtx plays with everything
					if (mpVolMesh->mbIntVtx[i] || !mpVolMesh->mbIntVtx[near_id])
					//if (mIntVtx[i]<=mIntVtx[near_id] || 
					//	(mIntVtx[i]==VTPt && mIntVtx[near_id]==VTEdge));
					if(is_visited[near_id] == 0 )
					{
						is_visited[near_id] = 1;
						near_list.push_back(near_id);
						//lNeighbour.insert(near_id);
						if (mIgnoreLapInput.count(near_id))
							mIgnoreLap.insert(i);
					}
				}
			}

			if (level==0)
			{
				vector<int> list_copy = near_list;
				list_copy.erase(list_copy.begin());
				oLap.laplacian_matrix.push_back(list_copy);
			}
		}
		near_list.erase(near_list.begin());
		//lap.laplacian_matrix[i] = near_list;
		//oLap.laplacian_matrix.push_back(near_list);
		oLap.mTransformNeighbourVtx.push_back(near_list);
	}

	oLap.point_list.clear();
#ifdef _EXT_SIM_LAP_ONLY
	assert(mpVolMesh->mSurfID2ID.size()>0);
	for (int i=0; i<(int)mpVolMesh->mSurfID2ID.size(); i++)
		oLap.point_list.push_back(trimesh::vec3(mXRef[mpVolMesh->mSurfID2ID[i]]));
	assert(oLap.point_list.size()==oLap.laplacian_matrix.size());

	for (int i=0; i<(int)oLap.laplacian_matrix.size(); i++)
	{
		vector<int>& near_list = oLap.laplacian_matrix[i];
		for (int j=0; j<(int)near_list.size(); j++)
		{
			near_list[j] = mpVolMesh->mID2SurfID[near_list[j]];
			assert(near_list[j]>=0 && near_list[j]<mpVolMesh->mSurfID2ID.size());
		}
	}
#else
	for (int i=0; i<(int)mXRef.size(); i++)
		oLap.point_list.push_back(trimesh::Vec<3, double>(mXRef[i]));
#endif

}

void PhysImpSolve3D::ComputeLapMat(int iDeform, _T_FLOAT iWeight)
{
	LaplacianDeformation lap;

	// build connections

	ComputeNRingNeighbour(mLapSimNumRing, lap);

	//if (mIgnoreLap.size()>0)
	//{
	//	vector<int> outSet;
	//	set_intersection(mIgnoreLap.begin(), mIgnoreLap.end(),
	//		lNeighbour.begin(), lNeighbour.end(),
	//		back_inserter(outSet));
	//}
	

	lap.calculateLaplacianvector();
	float average_near = 0;
	for(size_t i = 0; i < lap.laplacian_matrix.size(); i++)
	{
		average_near += lap.laplacian_matrix[i].size();
	}
	cout << "average neighbor size is " << average_near/lap.laplacian_matrix.size() << endl;

	// compute weight
	vector<_T_FLOAT> lapWeight(mXRef.size(), 1);
	//ComputeVtxWeightInverseVM(lapWeight);
	//ComputeVtxWeightNormalDiff(lapWeight);
	//ComputeVtxWeightZeroFreePointLap(lapWeight);
	if (mbMassLapWeight)
	{
		cout<<"Doing Relative Mass Weight\n";
		ComputeVtxWeightRelativeMass(lapWeight);
		EncodeVtxWeight(lapWeight, lap);
	}
	//ComputeVtxWeightLapDiff(lapWeight, lap);
	if (mbInvLapWeight)
	{
		cout<<"Doing Inv Lap Weight\n";
		ComputeVtxWeightInverseLap(lapWeight, lap);
	}

	// build matrix
	std::vector<std::map<int, double> > lapMat;
	std::vector<std::vector<double> > lapB;
	//ComputeNRingNeighbour(1, lap);
#ifdef _CURVATURE_WEIGHT
	lap.buildDeformationMatrix(0, 1e-6, lapMat, &lapB, &mIgnoreLap);
#else
	lap.buildDeformationMatrix(0, -1, lapMat, &lapB, &mIgnoreLapInput);
	//lap.buildDeformationMatrix(1, -1, lapMat, NULL);
#endif
			// for debug
	//{
	//	ofstream debugFileA[4];
	//	ofstream debugFileB("debugB.txt");
	//	//int interestPts[4] = {0, 26, 78, 90};
	//	// cube
	//	int interestPts[4] = {11, 0, 21, 1};
	//	//int interestPts[4] = {4, 28, 49, 261};
	//	for (int i=0; i<4; i++)
	//	{
	//		char nameBuffer[1000];
	//		sprintf_s<1000>(nameBuffer, "debugLapA%d.txt", i);
	//		debugFileA[i].open(nameBuffer);
	//	}

	//	//for (int i=0; i<4; i++)
	//	//{
	//	//	int idx = interestPts[i];
	//	//	for (int j=0; j<(int)mpVolMesh->mVtx.size(); j++)
	//	//	//if (j==idx)
	//	//	for (set<int>::iterator ite = mpVolMesh->mVtxConnect[j].begin();
	//	//		ite != mpVolMesh->mVtxConnect[j].end();
	//	//		ite++)
	//	//	if (*ite==idx)
	//	//		debugFileA[i]<<j<<":"<<mpVolMesh->mbIntVtx[j]<<"\t";
	//	//		//debugFileA[i]<<lap.laplacian_matrix[j][k]<<"\t";
	//	//}

	//	//for (int i=0; i<4; i++)
	//	//{
	//	//	int idx = interestPts[i];
	//	//	for (int j=0; j<(int)lap.laplacian_matrix.size(); j++)
	//	//	//if (j==idx)
	//	//	for (int k=0; k<(int)lap.laplacian_matrix[j].size(); k++)
	//	//	if (lap.laplacian_matrix[j][k]==idx)
	//	//		debugFileA[i]<<j<<"\t";
	//	//		//debugFileA[i]<<lap.laplacian_matrix[j][k]<<"\t";
	//	//}

	//	for (int i=0; i<4; i++)
	//	{
	//		int idx = interestPts[i];

	//		for (int c=0; c<3; c++)
	//			debugFileA[i]<<lap.vertex_laplacian_vector[idx][c]<<"\t";
	//		debugFileA[i]<<endl;
	//	}

	//	for (int i=0; i<4; i++)
	//	{
	//		int idx = interestPts[i]*3;
	//		for (map<int, double>::iterator ite = lapMat[idx].begin();
	//			ite != lapMat[idx].end();
	//			ite++)
	//			debugFileA[i]<<ite->second<<"\n";
	//	}

	//	for (int i=0; i<4; i++)
	//	{
	//		int idx = interestPts[i]*3;
	//		for (int j=0; j<(int)lapMat.size(); j++)
	//		for (map<int, double>::iterator ite = lapMat[j].begin();
	//			ite != lapMat[j].end();
	//			ite++)
	//		if (ite->first==idx)
	//			debugFileA[i]<<ite->second<<"\n";
	//	}
	//}


	// translate data
	//mLapSimMat.clear();

	if (mLapSimB.rows()==0)
		mLapSimB = MatXFloat::Zero(mX0.size()*3, 1);
	assert(lapB[0].size()==mLapSimB.rows());
	for (int i=0; i<(int)lapB[0].size(); i++)
		mLapSimB(i, 0) += lapB[0][i]*iWeight;

	std::map<SSSLib::Vec2i, _T_FLOAT> cmatTrip;
	for (int i=0; i<(int)lapMat.size(); i++)
	{
		for (map<int, double>::iterator ite = lapMat[i].begin();
			ite != lapMat[i].end();
			ite++)
		{
			Vec2i pst(i, ite->first);
			//Vec2i pst(ite->first, i);
			if (fabs(ite->second)!=0)
				cmatTrip[pst] = cmatTrip[pst]+ite->second;
		}
	}

	typedef Eigen::Triplet<_T_FLOAT> T;
	std::vector<T> tripleList;

	PushTriplet(cmatTrip, 1, tripleList);

	Eigen::SparseMatrix<_T_FLOAT> cMat(lapMat.size(), lap.point_list.size()*3);
	cMat.setFromTriplets(tripleList.begin(), tripleList.end());
	cMat.makeCompressed();

	Eigen::SparseMatrix<_T_FLOAT> cMatTranspose = Eigen::SparseMatrix<_T_FLOAT>(cMat.transpose());
	cMatTranspose.makeCompressed();

	if (mLapSimMatEigen.rows()==0)
	{
		mLapSimMatEigen = Eigen::SparseMatrix<_T_FLOAT>(mX0.size()*3, mX0.size()*3);
		mLapSimMatEigen.setZero();
	}
	mLapSimMatEigen += Eigen::SparseMatrix<_T_FLOAT>(cMatTranspose*cMat)*iWeight;
	mLapSimMatEigen.makeCompressed();

	//mLapHalfSimMatEigen += Eigen::SparseMatrix<_T_FLOAT>(cMat);

	mLapSimMat.clear();
	for (int k=0; k<mLapSimMatEigen.outerSize(); k++)
	for (Eigen::SparseMatrix<_T_FLOAT>::InnerIterator ite(mLapSimMatEigen, k);
		ite; ++ite)
#ifdef _EXT_SIM_LAP_ONLY
		mLapSimMat[Vec2i(
					mpVolMesh->mSurfID2ID[ite.row()/3]*3+ite.row()%3,
					mpVolMesh->mSurfID2ID[ite.col()/3]*3+ite.col()%3 )]
				+= ite.value();
#else
		mLapSimMat[Vec2i(ite.row(), ite.col())] += ite.value();
#endif
	
	mLapHalfSimMatEigen = cMat*sqrt(iWeight);

#ifdef _EXT_SIM_LAP_ONLY
	mLapSimMatEigen = Eigen::SparseMatrix<_T_FLOAT>(mXRef.size()*3, mXRef.size()*3);
	std::vector<T> newTriplets;
	PushTriplet(mLapSimMat, 1, newTriplets);
	mLapSimMatEigen.setFromTriplets(newTriplets.begin(), newTriplets.end());
	mLapSimMatEigen.makeCompressed();
	for (int k=0; k<cMat.outerSize(); k++)
	{
		std::vector<T> newHalfTriplets;
		for (Eigen::SparseMatrix<_T_FLOAT>::InnerIterator ite(cMat, k);
		ite; ite++)
			newHalfTriplets.push_back(T(mpVolMesh->mSurfID2ID[ite.row()/3]*3+ite.row()%3,
					mpVolMesh->mSurfID2ID[ite.col()/3]*3+ite.col()%3,
					ite.value()));
		mLapHalfSimMatEigen = Eigen::SparseMatrix<_T_FLOAT>(mXRef.size()*3, mXRef.size()*3);
		mLapHalfSimMatEigen.setFromTriplets(newHalfTriplets.begin(), newHalfTriplets.end());
	}
#endif
	//mLapSimMatEigen = Eigen::SparseMatrix<_T_FLOAT>(mXRef.size()*3, mXRef.size()*3);
	//mLapSimMatEigen.setFromTriplets(tripleList.begin(), tripleList.end());
	//mLapSimMatEigen.makeCompressed();
}

void PhysImpSolve3D::DecodeHalfLapMat(std::vector<Eigen::Triplet<_T_FLOAT> >& iList)
{
	mLapHalfSimMatEigen = Eigen::SparseMatrix<_T_FLOAT>(mXRef.size()*3,
												mXRef.size()*3);
	mLapHalfSimMatEigen.setFromTriplets(iList.begin(), iList.end());
	mLapHalfSimMatEigen.makeCompressed();

	Eigen::SparseMatrix<_T_FLOAT> halfTrans(mLapHalfSimMatEigen.transpose());
	mLapSimMatEigen = halfTrans*mLapHalfSimMatEigen;

	mLapSimMat.clear();
	for (int k=0; k<mLapSimMatEigen.outerSize(); k++)
	for (Eigen::SparseMatrix<_T_FLOAT>::InnerIterator ite(mLapSimMatEigen, k);
		ite; ++ite)
		mLapSimMat[Vec2i(ite.row(), ite.col())] = ite.value();
}

void PhysImpSolve3D::ComputeOneRingSimMat(int iDeform)
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	std::vector<T> tripleList;
	LaplacianDeformation lap;
	ComputeNRingNeighbour(1, lap);
	lap.calculateLaplacianvector();
	vector<_T_FLOAT> lapWeight;
	//ComputeVtxWeightInverseVM(lapWeight);
	//ComputeVtxWeightNormalDiff(lapWeight);
	//ComputeVtxWeightZeroFreePointLap(lapWeight);
	//EncodeVtxWeight(lapWeight, lap);
	//ComputeVtxWeightLapDiff(lapWeight, lap);
#ifdef _EXT_SIM_LAP_ONLY
	//assert(lapWeight.size()==mpVolMesh->mSurfID2ID.size());
#endif

	for (int i=0; i<(int)mXRef.size(); i++)
	if (mIntVtx[i] != VTInt)
	{
		set<int>& neighbourList =  mpVolMesh->mVtxConnect[i];
		vector<int> elements;
		elements.push_back(i);
		for (set<int>::iterator ite = neighbourList.begin();
			ite != neighbourList.end();
			ite++)
		if (mIntVtx[*ite] != VTInt)
			elements.push_back(*ite);

		_T_FLOAT weight = 1;
		if (lapWeight.size()>0)
#ifdef _EXT_SIM_LAP_ONLY
			weight = lapWeight[mpVolMesh->mID2SurfID[i]];
#else
			weight = lapWeight[i];
#endif

		AddSimMatElement(iDeform,
						weight,
						elements,
						tripleList);
	}

	DecodeHalfLapMat(tripleList);
}

void PhysImpSolve3D::ComputeElementSimMat(int iDeform)
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	std::vector<T> tripleList;

	for (int eleType=0; eleType<2; eleType++)
	{
		int eleNum;
		int eleSize;
		switch (eleType)
		{
		case 0:
			eleNum = mpVolMesh->mTets.size();
			eleSize = 4;
			break;
		case 1:
			eleNum = mpVolMesh->mHexs.size();
			eleSize = 8;
			break;
		}
		for (int t=0; t<eleNum; t++)
		{
			vector<int> ele(eleSize);
			if (eleType==0)
				for (int i=0; i<eleSize; i++)
					ele[i] = mpVolMesh->mTets[t][i];
			else
				for (int i=0; i<eleSize; i++)
					ele[i] = mpVolMesh->mHexs[t][i];

			AddSimMatElement(iDeform,
							1.0,
							ele,
							tripleList);
		}
	}

	DecodeHalfLapMat(tripleList);
}

void PhysImpSolve3D::AddSimMatElement(int iDeform,
									_T_FLOAT iWeight,
									std::vector<int>& iElements,
									std::vector<Eigen::Triplet<_T_FLOAT> >& oList)
{
	Eigen::MatrixXd CMatrix;
	if (iDeform==LaplacianDeformation::DEFORM_AFFINE)
	{
		CMatrix = Eigen::MatrixXd::Constant(3*iElements.size(), 12, 0);
		for(int j = 0; j <iElements.size(); j++)
		{
			int index_j = iElements[j];
			SVec3Float pos_j = mXRef[index_j];

			for (int r=0; r<3; r++)
			{
				for (int c=0; c<3; c++)
				{
					CMatrix(j*3+r, r*3+c) = pos_j[c];
				}
				CMatrix(j*3+r, 9+r) = 1;
			}
		}
	}
	else
	{
		cout<<"Feature not implemented!";
		exit(1);
	}


	Eigen::MatrixXd CTranspose(CMatrix.transpose());
	Eigen::MatrixXd CTC = CTranspose*CMatrix;
	//Eigen::MatrixXd CTCI = CTC.inverse();
	Eigen::MatrixXd CTCI = LaplacianDeformation::pinv(CTC, 1e-8);
	Eigen::MatrixXd HMatrix = CMatrix*CTCI*CTranspose;

	HMatrix -= Eigen::MatrixXd::Identity(HMatrix.rows(), HMatrix.cols());
	assert(HMatrix.rows()==iElements.size()*3);
	assert(HMatrix.cols()==iElements.size()*3);
	for (int r=0; r<iElements.size()*3; r++)
	for (int c=0; c<iElements.size()*3; c++)
		oList.push_back(Eigen::Triplet<_T_FLOAT>(iElements[r/3]*3+(r%3),
												iElements[c/3]*3+(c%3),
												HMatrix(r, c)));
}

_T_FLOAT PhysImpSolve3D::EvaluateLapSim()
{
	
	MatXFloat xvec(mX0.size()*3, 1);
	for (int i=0; i<(int)mX0.size(); i++)
	for (int c=0; c<3; c++)
		xvec(i*3+c, 0) = mX0[i][c];

	
	MatXFloat xt = xvec;
	xt.transposeInPlace();
	//cout<<"Lap Sim Mat Size:"<<mLapSimMatEigen.rows()<<","<<mLapSimMatEigen.cols()<<"\t"
	//	<<"xvec size:"<<xvec.rows()<<"\t";
	MatXFloat val = xt*mLapSimMatEigen*xvec;
	if (mLapSimB.rows()>0)
		val -= xt*mLapSimB*2;
	assert(val.rows()==1);
	assert(val.cols()==1);
	//assert(val(0, 0)>=-mEdgeLengthEpsilon*mX0.size());
	//if (val(0, 0)<0)
	//	cout<<"Warning: Negative Laplacian energy, possibly numerical error:"<<val(0, 0)<<endl;
	//return val(0, 0)*mLapSimWeight;
	return val(0, 0);
	
	/*Eigen::VectorXd halfVec = mLapHalfSimMatEigen*xvec;
	_T_FLOAT val = halfVec.squaredNorm();
	return val*mLapSimWeight;*/
}

void PhysImpSolve3D::LapAddToGradient(_T_FLOAT iScale, _T_FLOAT* oGrad)
{
	MatXFloat xvec(mX0.size()*3, 1);
	for (int i=0; i<(int)mX0.size(); i++)
	for (int c=0; c<3; c++)
		xvec(i*3+c, 0) = mX0[i][c];

	//MatXFloat halfMat(mLapHalfSimMatEigen);
	//halfMat.transposeInPlace();
	//MatXFloat val = halfMat*mLapHalfSimMatEigen*xvec;
	MatXFloat val = 2*mLapSimMatEigen*xvec;
	for (int i=0; i<val.rows(); i++)
		//oGrad[i] += val(i, 0)*iScale*mLapSimWeight;
		oGrad[i] += val(i, 0)*iScale;

	if (mLapSimB.rows()>0)
		for (int i=0; i<val.rows(); i++)
			oGrad[i] -= mLapSimB(i, 0)*iScale*2;
}

void PhysImpSolve3D::LapAddToKKT(_T_FLOAT iScale)
{
	Vec2i lastKey(-1, -1);
	KKTEle lastEle;
	for (map<Vec2i, _T_FLOAT>::iterator ite = mLapSimMat.begin();
		ite != mLapSimMat.end();
		ite++)
	{
		int idxJ = ite->first[1];

		assert(ite->first[0]>=lastKey[0]);
		if (ite->first[0]!=lastKey[0] && lastKey[0]!=-1)
		{
			mKKT.push_back(lastEle);
			lastEle = KKTEle();
			if (mLapSimB.rows()>0)
				lastEle.mB += mLapSimB(idxJ, 0);
		}
		
		lastEle.AddACoe(idxJ, ite->second*2);
		lastEle.mB += ite->second*mX0[idxJ/3][idxJ%3];

		lastKey = ite->first;
	}

	if (lastKey[0]!=-1)
		mKKT.push_back(lastEle);
}


