/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "Static2D.h"
#include <eigen/sparse>
#include <eigen/iterativeLinearSolvers>
#include <eigen/sparselu>
#include <iostream>
#include "../SSSLib/SSSFunc.h"

using namespace std;
using namespace Eigen;

Static2D::Static2D(void)
{
}

Static2D::~Static2D(void)
{
}

void Static2D::ComputeBoundaryVtx()
{
	mBoundaryVtx.resize(mRest.size());
	vector<int> adjTriangles(mRest.size(), 0);
	for (int t=0; t<(int)mTriangles.size(); t++)
	for (int v=0; v<3; v++)
		adjTriangles[mTriangles[t][v]]++;
	for (int i=0; i<(int)mRest.size(); i++)
	if (adjTriangles[i]<6)
		mBoundaryVtx[i] = true;
	else
		mBoundaryVtx[i] = false;
}

void Static2D::InitRect(int iWidth, int iHeight, _T_FLOAT iScale)
{
	mRest.resize(iWidth*iHeight);
	mForce.resize(iWidth*iHeight);
	mForce.assign(iWidth*iHeight, Vec2Float(0, 0));
	mFixedVtx.resize(mRest.size());
	mFixedVtx.assign(mFixedVtx.size(), 0);

	for (int r=0; r<iHeight; r++)
	for (int c=0; c<iWidth; c++)
		mRest[r*iWidth+c] = Vec2Float(_T_FLOAT(c)-(iWidth-1)*0.5,
			_T_FLOAT(r)-(iHeight-1)*0.5)*iScale;

	mTriangles.resize((iWidth-1)*(iHeight-1)*2);

	for (int r=0; r<iHeight-1; r++)
	for (int c=0; c<iWidth-1; c++)
	{
		mTriangles[(r*(iWidth-1)+c)*2][0] = r*iWidth+c;
		mTriangles[(r*(iWidth-1)+c)*2][1] = r*iWidth+c+1;
		mTriangles[(r*(iWidth-1)+c)*2][2] = (r+1)*iWidth+c+1;

		mTriangles[(r*(iWidth-1)+c)*2+1][0] = (r+1)*iWidth+c+1;
		mTriangles[(r*(iWidth-1)+c)*2+1][1] = (r+1)*iWidth+c;
		mTriangles[(r*(iWidth-1)+c)*2+1][2] = r*iWidth+c;

		/*mTriangles[(r*(iWidth-1)+c)*2+1] = Eigen::Vector3i(
			(r+1)*iWidth+c+1,
			(r+1)*iWidth+c,
			r*iWidth+c);*/
	}
}

void Static2D::InitParam(_T_FLOAT iYM, _T_FLOAT iPR)
{
	mD = MatXFloat::Zero(3, 3);
	mD(0, 0) = 1;
	mD(1, 1) = 1;
	mD(0, 1) = iPR;
	mD(1, 0) = iPR;
	mD(2, 2) = (1-iPR)*0.5;
	mD = mD*(iYM/(1+iPR*iPR));

	mThick = 0.1;
}

void Static2D::SetGravity(float iDensity)
{
	for (int i=0; i<(int)mTriangles.size(); i++)
	{
		SSSLib::Vec3i t = mTriangles[i];
		Vec2Float vtx[3];
		vtx[0] = mRest[t[0]];
		vtx[1] = mRest[t[1]];
		vtx[2] = mRest[t[2]];
		_T_FLOAT volume = SSSLib::cross2<_T_FLOAT>(vtx[1]-vtx[0], vtx[2]-vtx[0])*mThick;
		_T_FLOAT gravity = volume*iDensity*9.8;
		for (int i=0; i<3; i++)
			mForce[t[i]][1] -= gravity/3.0f;
	}
}

void Static2D::SetGravityWithMassVec(const std::vector<_T_FLOAT>& iMass)
{
	for (int i=0; i<(int)mRest.size(); i++)
	{
		mForce[i][1] =- 9.8*iMass[i];
	}
}

void Static2D::ComputeF(SVec2Float iX[3], SVec2Float iP[3], _T_FLOAT iScale, MatXFloat iD, SVec2Float* oF)
{
	_T_FLOAT x1 = iX[0][0];
	_T_FLOAT x2 = iX[1][0];
	_T_FLOAT x3 = iX[2][0];

	_T_FLOAT y1 = iX[0][1];
	_T_FLOAT y2 = iX[1][1];
	_T_FLOAT y3 = iX[2][1];

	_T_FLOAT detJ = (x1-x3)*(y2-y3)-(y1-y3)*(x2-x3);

	MatXFloat B = MatXFloat::Zero(3, 6);

	B(0, 0) = y2-y3;
	B(0, 2) = y3-y1;
	B(0, 4) = y1-y2;

	B(1, 1) = x3-x2;
	B(1, 3) = x1-x3;
	B(1, 5) = x2-x1;

	B(2, 0) = x3-x2;
	B(2, 1) = y2-y3;
	B(2, 2) = x1-x3;
	B(2, 3) = y3-y1;
	B(2, 4) = x2-x1;
	B(2, 5) = y1-y2;

	B /= detJ;
		
	MatXFloat k = B.transpose()*iD*B*iScale*0.5*fabs(detJ);
	MatXFloat q = MatXFloat::Zero(6, 1);
	for (int i=0; i<3; i++)
	{
		q(i*2, 0) = iP[i][0]-iX[i][0];
		q(i*2+1, 0) = iP[i][1]-iX[i][1];
	}
	MatXFloat f = k*q;
	for (int i=0; i<3; i++)
	{
		oF[i][0] = f(i*2,0);
		oF[i][1] = f(i*2+1,0);
	}
}

void Static2D::ComputeStress(SVec2Float iX[3], SVec2Float iP[3], MatXFloat iD, SVec3Float& oStress)
{
	_T_FLOAT x1 = iX[0][0];
	_T_FLOAT x2 = iX[1][0];
	_T_FLOAT x3 = iX[2][0];

	_T_FLOAT y1 = iX[0][1];
	_T_FLOAT y2 = iX[1][1];
	_T_FLOAT y3 = iX[2][1];

	_T_FLOAT detJ = (x1-x3)*(y2-y3)-(y1-y3)*(x2-x3);

	MatXFloat B = MatXFloat::Zero(3, 6);

	B(0, 0) = y2-y3;
	B(0, 2) = y3-y1;
	B(0, 4) = y1-y2;

	B(1, 1) = x3-x2;
	B(1, 3) = x1-x3;
	B(1, 5) = x2-x1;

	B(2, 0) = x3-x2;
	B(2, 1) = y2-y3;
	B(2, 2) = x1-x3;
	B(2, 3) = y3-y1;
	B(2, 4) = x2-x1;
	B(2, 5) = y1-y2;

	B /= detJ;
	MatXFloat q = MatXFloat::Zero(6, 1);
	for (int i=0; i<3; i++)
	{
		q(i*2, 0) = iP[i][0]-iX[i][0];
		q(i*2+1, 0) = iP[i][1]-iX[i][1];
	}

	MatXFloat stress = iD*B*q;
	oStress[0] = stress(0, 0);
	oStress[1] = stress(1, 0);
	oStress[2] = stress(2, 0);
}

void Static2D::Compute()
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	std::vector<T> tripleList;

	for (int tri=0; tri<(int)mTriangles.size(); tri++)
	{
		_T_FLOAT x1 = mRest[mTriangles[tri][0]](0);
		_T_FLOAT x2 = mRest[mTriangles[tri][1]](0);
		_T_FLOAT x3 = mRest[mTriangles[tri][2]](0);

		_T_FLOAT y1 = mRest[mTriangles[tri][0]](1);
		_T_FLOAT y2 = mRest[mTriangles[tri][1]](1);
		_T_FLOAT y3 = mRest[mTriangles[tri][2]](1);

		_T_FLOAT detJ = (x1-x3)*(y2-y3)-(y1-y3)*(x2-x3);

		MatXFloat B = MatXFloat::Zero(3, 6);

		B(0, 0) = y2-y3;
		B(0, 2) = y3-y1;
		B(0, 4) = y1-y2;

		B(1, 1) = x3-x2;
		B(1, 3) = x1-x3;
		B(1, 5) = x2-x1;

		B(2, 0) = x3-x2;
		B(2, 1) = y2-y3;
		B(2, 2) = x1-x3;
		B(2, 3) = y3-y1;
		B(2, 4) = x2-x1;
		B(2, 5) = y1-y2;

		B /= detJ;
		
		MatXFloat k = B.transpose()*mD*B*mThick*0.5*fabs(detJ);

		for (int r=0; r<6; r++)
		for (int c=0; c<6; c++)
			tripleList.push_back(
				T(
					mTriangles[tri][r/2]*2+r%2,
					mTriangles[tri][c/2]*2+c%2,
					k(r,c)
				));
	}


	int ASize = 2*(int)mRest.size();
	for (int i=0; i<(int)mRest.size(); i++)
	if (mFixedVtx[i])
	{
		tripleList.push_back(T(ASize+0, i*2+0, 1));
		tripleList.push_back(T(ASize+1, i*2+1, 1));

		tripleList.push_back(T(i*2+0, ASize+0, 1));
		tripleList.push_back(T(i*2+1, ASize+1, 1));

		ASize += 2;
	}

	SparseMatrix<_T_FLOAT> A(ASize, ASize);
	A.setFromTriplets(tripleList.begin(), tripleList.end());

	MatXFloat b = MatXFloat::Zero(ASize, 1);
	for (int i=0; i<(int)mForce.size(); i++)
	{
		b(i*2+0) = mForce[i](0);
		b(i*2+1) = mForce[i](1);
	}

	SparseLU<SparseMatrix<_T_FLOAT>, COLAMDOrdering<int> > solver;
	//BiCGSTAB<SparseMatrix<_T_FLOAT> > solver;
	solver.analyzePattern(A);
	solver.factorize(A);
	if (solver.lastErrorMessage().length()>0)
		cout<<solver.lastErrorMessage();
	MatXFloat x = solver.solve(b);

	mStatic.resize(mRest.size());
	for (int i=0; i<(int)mStatic.size(); i++)
	{
		mStatic[i] = Vec2Float(x(i*2), x(i*2+1))+mRest[i];

		//cout<<i<<":("<<mStatic[i](0)<<","<<mStatic[i](1)<<")"<<endl;
	}
}

void Static2D::ComputeStrainStress()
{
	mTriangleStrain.resize(mTriangles.size());
	mTriangleStress.resize(mTriangles.size());
	mTrianglePS.resize(mTriangles.size());
	mTriangleNormalizedPS.resize(mTriangles.size());

	for (int f=0; f<(int)mTriangles.size(); f++)
	{
		SSSLib::Vec3i face = mTriangles[f];
		Vec2Float x1 = mRest[face[1]]-mRest[face[0]];
		Vec2Float x2 = mRest[face[2]]-mRest[face[0]];
		Vec2Float p1 = mStatic[face[1]]-mStatic[face[0]];
		Vec2Float p2 = mStatic[face[2]]-mStatic[face[0]];

		Mat2x2 x;
		x(0, 0) = x1[0];
		x(1, 0) = x1[1];
		x(0, 1) = x2[0];
		x(1, 1) = x2[1];
		Mat2x2 p;
		p(0, 0) = p1[0];
		p(1, 0) = p1[1];
		p(0, 1) = p2[0];
		p(1, 1) = p2[1];
		Mat2x2 strain = p*x.inverse()-Mat2x2::Identity();

		strain = 0.5*(strain+strain.transpose());
		Vec3Float strainVec(strain(0, 0), strain(1,1), strain(0, 1));
		Vec3Float stressVec = mD * strainVec;
		Mat2x2 stress;
		stress(0, 0) = stressVec[0];
		stress(0, 1) = stress(1, 0) = stressVec[2];
		stress(1, 1) = stressVec[1];

		mTriangleStrain[f] = strain;
		mTriangleStress[f] = stress;

		Eigen::Matrix<std::complex<double>, 2, 1> eigval = stress.eigenvalues();
		mTrianglePS[f][0] = eigval(0, 0).real();
		mTrianglePS[f][1] = eigval(1, 0).real();
		if (fabs(mTrianglePS[f][1])>fabs(mTrianglePS[f][0]))
			swap(mTrianglePS[f][0], mTrianglePS[f][1]);
	}

	mVtxStrain.clear();
	mVtxStress.clear();
	mVtxPS.clear();
	mVtxStrain.resize(mStatic.size(), Mat2x2::Zero());
	mVtxStress.resize(mStatic.size(), Mat2x2::Zero());
	mVtxPS.resize(mStatic.size(), Vec2Float::Zero());
	mVtxNormalizedPS.resize(mStatic.size());

	vector<_T_FLOAT> weight(mStatic.size(), 0);
	for (int f=0; f<(int)mTriangles.size(); f++)
	{
		SSSLib::Vec3i face = mTriangles[f];
		for (int i=0; i<3; i++)
		{
			mVtxStrain[face[i]] += mTriangleStrain[f];
			mVtxStress[face[i]] += mTriangleStress[f];
			weight[face[i]] += 1.0;
		}
	}

	for (int v=0; v<(int)mVtxStrain.size(); v++)
	{
		mVtxStrain[v] /= weight[v];
		mVtxStress[v] /= weight[v];
		Eigen::Matrix<std::complex<double>, 2, 1> eigval = mVtxStress[v].eigenvalues();
		mVtxPS[v][0] = eigval(0, 0).real();
		mVtxPS[v][1] = eigval(1, 0).real();
		if (fabs(mVtxPS[v][1])>fabs(mVtxPS[v][0]))
			swap(mVtxPS[v][0], mVtxPS[v][1]);
	}

	_T_FLOAT maxPS = mTrianglePS[0][0];
	_T_FLOAT minPS = mTrianglePS[0][0];
	for (int i=0; i<(int)mTrianglePS.size(); i++)
	{
		maxPS = SSSLib::Max(maxPS, mTrianglePS[i][0]);
		minPS = SSSLib::Min(minPS, mTrianglePS[i][0]);
	}

	for (int i=0; i<(int)mTrianglePS.size(); i++)
	if (maxPS!=minPS)
		mTriangleNormalizedPS[i] = (mTrianglePS[i][0]-minPS)/(maxPS-minPS);
	else
		mTriangleNormalizedPS[i] = 1.0;

	for (int v=0; v<(int)mVtxStrain.size(); v++)
	if (maxPS!=minPS)
		mVtxNormalizedPS[v] = (mVtxPS[v][0]-minPS)/(maxPS-minPS);
	else
		mVtxNormalizedPS[v] = 1.0;

	mMinPS = minPS;
	mMaxPS = maxPS;
}

Vec2Float Static2D::SamplePS(const Vec2Float& iPst, int iTriID)
{
	SSSLib::Vec3i tri = mTriangles[iTriID];
	Vec2Float vec[3];
	for (int v=0; v<3; v++)
		vec[v] = mRest[tri[v]];

	Vec2Float ps(0, 0);
	_T_FLOAT w[3];
	for (int v=0; v<3; v++)
	{
		w[v] = SSSLib::cross2<_T_FLOAT>(vec[(v+1)%3]-iPst,vec[(v+2)%3]-iPst);
		ps += w[v]*mVtxPS[tri[v]];
	}

	return ps/(w[0]+w[1]+w[2]);
}